# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import Counter
import os
import re

from .BufferPool import BufferPool
from .Connection import Connection
from .LocalVars import LocalVars
from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject
from .HIrReceive import HIrReceive
from .HeavyLangObject import HeavyLangObject


class HeavyGraph(HeavyIrObject):
    """ Represents a graph. Subclasses HeavyIrObject for functionality.
        "__graph" does exist as a HeavyIR object, though it has no functionality
        and should not appear in any IR output.
    """

    def __init__(self, graph=None, graph_args=None, file=None, xname=None):
        # zero inlets and outlets until inlet/outlet objects are declared
        HeavyIrObject.__init__(self, "__graph", graph_args, graph, 0, 0)

        # the heavy file which defines this graph
        self.file = file

        # a user-defined name of this graph
        self.xname = xname or "heavy"

        # the dictionary of all objects in the graph
        self.objs = {}

        # set the local arguments
        self.args = graph_args or {}

        # initialise the local variables
        self.local_vars = LocalVars()

        # the list of all constituent inlet and outlet objects
        # graphs always start with no inlet/outlets
        self.inlet_objs = []
        self.outlet_objs = []

        # the set of input/output channels that this graph writes to
        self.input_channel_set = set()
        self.output_channel_set = set()

        # an ordered list of signal objects to process
        self.signal_order = []

        # a pool of signal buffers for use during signal ordering and buffer assignment
        self.buffer_pool = None

    def resolve_arguments(self, obj_args):
        """ Resolves the object arguments based on values from the local graph.
        """
        args = dict(obj_args)  # make a copy of the input arguments
        for key, value in args.items():
            # do any of them reference a variable input?
            if isinstance(value, str):
                if value.find("$") > -1:
                    # is there a value provided for that key in the graph arguments?
                    for k in self.args:
                        # replace all instances of $k with the argument value
                        # do this for all keys in the graph's argument dictionary
                        dollar_key = "$" + k
                        if value.find(dollar_key) > -1:
                            if value == dollar_key:
                                value = self.args[k]  # maintain the original data type
                                break  # break because all dollar arguments have by definition been replaced
                            else:
                                # otherwise assume a string value
                                value = value.replace(dollar_key, str(self.args[k]))
                                # continue around the loop in case value has multiple dollar args

                    if isinstance(value, str) and value.find("$") > -1:
                        # if the variable cannot be resolved, leave it alone.
                        # When the arguments are passed to the object, they will be
                        # set to the default value. These unset variables are simply
                        # removed from the argument dictionary
                        args.pop(key, None)
                        self.add_warning("Variable \"{0}\":\"{1}\" could not be resolved.".format(key, value))
                    else:
                        args[key] = value
        return args

    def connect_objects(self, c, require_intra_graph_connection=True):
        """ Establish a connection from one object to another.
            Typically a connection must be made within a single graph, though
            this requirement may be subpressed.
        """
        if require_intra_graph_connection:
            assert c.from_object.graph is c.to_object.graph, \
                "Conection established between two objects not in the same graph: {0}".format(c)

        # add the connection from the parent to the child
        c.from_object.add_connection(c)

        # add the connection to the child from the parent
        c.to_object.add_connection(c)

    def disconnect_objects(self, c):
        """ Remove a connection from one object to another.
        """
        # remove the connection from the parent to the child
        c.from_object.remove_connection(c)

        # remove the connection to the child from the parent
        c.to_object.remove_connection(c)

    def update_connection(self, c, n_list):
        """ Update the old connection (c) with the new connections (n_list).
            Connection order is maintained. If c is None, this method acts
            as connect_objects(). If n is empty, this method acts as disconnect_objects().
        """
        if c is not None and len(n_list) > 0:
            if all([c.from_object is n.from_object for n in n_list]):
                # connections start at the same point
                c.from_object.replace_connection(c, n_list)
                c.to_object.remove_connection(c)
                for n in n_list:
                    n.to_object.add_connection(n)
            elif all([c.to_object is n.to_object for n in n_list]):
                # connections end at the same point
                c.to_object.replace_connection(c, n_list)
                c.from_object.remove_connection(c)
                for n in n_list:
                    n.from_object.add_connection(n)
            else:
                raise HeavyException("Connections must have a common endpoint: {0} / {1}".format(c, n))
        elif c is not None and len(n_list) == 0:
            self.disconnect_objects(c)  # remove connection c
        elif c is None and len(n_list) > 0:
            for n in n_list:
                self.connect_objects(n)  # add connection n

    def add_object(self, obj, obj_id=None):
        """ Add an object to the graph based on the given id.
            Per-object bookkeeping is performed.
        """

        obj_id = obj_id or obj.id
        if obj_id not in self.objs:
            obj.graph = self
            self.objs[obj_id] = obj

            # some object needs to be specially handled when added to the graph
            if obj.type in ["inlet", "__inlet"]:
                self.inlet_connections.append([])
                self.inlet_buffers.append(("zero", 0))
                self.inlet_objs.append(obj)
                # sort inlet objects according to their index, ascending
                self.inlet_objs.sort(key=lambda o: o.args["index"])
            elif obj.type in ["outlet", "__outlet"]:
                self.outlet_connections.append([])
                self.outlet_buffers.append(("zero", 0))
                self.outlet_objs.append(obj)
                self.outlet_objs.sort(key=lambda o: o.args["index"])
            elif obj.type in ["adc"]:
                self.input_channel_set.update(obj.args["channels"])
            elif obj.type in ["dac"]:
                # if the object is a dac, keep track of the output channels
                # that this graph is writing to
                self.output_channel_set.update(obj.args["channels"])
            elif obj.type in ["receive", "__receive", "send", "__send"]:
                self.__register_named_object(obj, obj.name)
            elif obj.type in ["table", "__table"]:
                self.__register_named_object(
                    obj,
                    obj.name,
                    static=obj.static,
                    unique=True)
            elif obj.type in ["var"] and obj.name is not None:
                self.__register_named_object(
                    obj,
                    obj.name,
                    static=obj.static,
                    unique=True)
        else:
            self.add_error("Duplicate object id {0} found in graph.".format(obj_id))

    def __register_named_object(self, obj, name=None, static=False, unique=False):
        """ Register a named object at the appropriate graph level.
        """

        name = name or obj.name
        if obj.scope == "private" and not static:
            # stay at this level
            self.local_vars.register_object(obj, name, static, unique)
        elif obj.scope == "protected":
            # go up one level
            g = self.graph if self.graph is not None else self
            g.local_vars.register_object(obj, name, static, unique)
        elif obj.scope == "public" or static:
            self.get_root_graph().local_vars.register_object(obj, name, static, unique)
        else:
            raise HeavyException("Unknown scope \"{0}\" for object {1}.".format(obj.scope, obj))

    def __unregister_named_object(self, obj, name=None, static=False, unique=False):
        """ Unregister a named object at the appropriate graph level.
        """

        name = name or obj.name
        if obj.scope == "private" and not static:
            self.local_vars.unregister_object(obj, name)
        elif obj.scope == "protected":
            # go up one level
            g = self.graph if self.graph is not None else self
            g.local_vars.unregister_object(obj, name)
        elif obj.scope == "public" or static:
            self.get_root_graph().local_vars.unregister_object(obj, name)
        else:
            raise HeavyException("Unknown scope \"{0}\" for object {1}.".format(obj.scope, obj))

    def resolve_objects_for_name(self, name, obj_types, local_graph=None):
        """ Returns all objects with the given name and type that are visible
            from this graph. Returns an empty list if no objects could be found.
            The results are otherwise ordered from most local/private first,
            and the most global last.
        """

        obj_types = obj_types if isinstance(obj_types, list) else [obj_types]
        local_graph = local_graph or self

        obj_list = [o for o in self.local_vars.get_objects_for_name(name, obj_types)
                    if not (o.scope == "private" and o.graph is not local_graph)]

        if self.graph is not None:
            obj_list_from_super_graph = self.graph.resolve_objects_for_name(name, obj_types, local_graph)
            obj_list.extend(obj_list_from_super_graph)

        return obj_list

    def resolve_object_for_name(self, name, obj_types, local_graph=None):
        """ Returns the first object with the given name and type that is visible
            from this graph. Returns None if no objects are available.
            This is a convenience method.
        """

        objs = self.resolve_objects_for_name(name, obj_types, local_graph)
        return objs[0] if len(objs) > 0 else None

    def is_root_graph(self):
        """ Returns true if this is the top-level (i.e. root) graph. False otherwise.
        """
        return self.graph is None

    def get_root_graph(self):
        """Returns the top-level graph.
        """
        return self if self.is_root_graph() else self.graph.get_root_graph()

    def find_path_for_abstraction(self, obj_type):
        """ Travels up the graph heirarchy looking for a file path to an abstraction.
            Returns None if no abstraction is found.
        """
        file_path = self.local_vars.find_path_for_abstraction(obj_type)
        if file_path is not None:
            return file_path
        else:
            return self.graph.find_path_for_abstraction(obj_type) if self.graph is not None else None

    def remove_object(self, o, obj_id=None):
        """ Removes an object and all of its connections from the graph.
            A custom id for the object to be removed can be given.
        """
        for connections in o.inlet_connections:
            for c in list(connections):  # make copy of connections as it will be mutated
                self.disconnect_objects(c)
        for connections in o.outlet_connections:
            for c in list(connections):  # make copy
                self.disconnect_objects(c)

        for k, v in self.objs.items():
            if o is v:
                del self.objs[k]
                break

        if o.type in ["receive", "__receive", "send", "__send"]:
            self.__unregister_named_object(o, o.name)

    def get_inlet_object(self, index):
        """ Returns the indexed inlet object of this graph.
        """
        return self.inlet_objs[index]

    def get_outlet_object(self, index):
        """ Returns the indexed outlet object of this graph.
        """
        return self.outlet_objs[index]

    def get_input_channel_set(self, recursive=False):
        """ Returns the set of input channels that this graph writes to.
            Optionally includes all subgraphs as well.
        """
        if recursive:
            channels = set(self.input_channel_set)  # copy the output channel set
            for o in [o for o in self.objs.values() if o.type == "__graph"]:
                channels.update(o.get_input_channel_set(recursive=True))
            return channels
        else:
            return self.input_channel_set

    def get_output_channel_set(self, recursive=False):
        """ Returns the set of output channels that this graph writes to.
            Optionally includes all subgraphs as well.
        """
        if recursive:
            channels = set(self.output_channel_set)  # copy the output channel set
            for o in [o for o in self.objs.values() if o.type == "__graph"]:
                channels.update(o.get_output_channel_set(recursive=True))
            return channels
        else:
            return self.output_channel_set

    def get_notices(self):
        notices = HeavyLangObject.get_notices(self)
        for o in self.objs.values():
            n = o.get_notices()
            notices["warnings"].extend(n["warnings"])
            notices["errors"].extend(n["errors"])
        return notices

    def get_objects_for_type(self, obj_type, recursive=False):
        """ Returns a list of all objects of a given type in this graph.
            The optional parameter "recursive" also includes all objects from subgraphs.
        """
        obj_list = []
        for o in self.objs.values():
            if o.type == "__graph" and recursive:
                obj_list.extend(o.get_objects_for_type(obj_type, recursive))
            elif o.type == obj_type:
                obj_list.append(o)
        return obj_list

    def get_object_counter(self, recursive=False):
        """ Returns a counter of all object types in this graph.
            Graph objects are explicitly removed. "__inlet" and "__outlet"
            objects are renamed to "inlet" and "outlet", for clarity.
            An optional recursive argument includes all subgraphs.
        """
        c = Counter()
        for o in self.objs.values():
            if o.type == "__graph" and recursive:
                c += o.get_object_counter(recursive=True)
            elif o.type in ["__inlet", "__outlet"]:
                c[o.type[2:]] += 1  # remove leading "__"
            else:
                c[o.type] += 1
        return c

    def prepare(self):
        """ Prepares a graph to be exported. Must be called from a root graph.
        """
        assert self.is_root_graph()

        try:
            # apply graph transformations when all graphs have been read
            # All transformations must be recursive.

            # resolve all connection types such that all HeavyLang objects can
            # be correctly reduced
            self._resolve_connection_types()

            # TODO(mhroth): prune unnecessary objects and connections
            # All unnecessary objects and connections should be removed before the
            # graph is reduced. Reduction generally takes into account the existence
            # of connections when deciding on what HeavyIR objects to used.
            self._remove_unused_inlet_connections()

            # reconnect all send~/receive~ objects and remove them from the graph
            self.remap_send_receive()

            # now that the basic graph has been constructed,
            # objects are reduced to their well-defined low-level types.
            # Some objects may be replaced by graphs containing only low-level objects.
            # The graph is changed in-place and consists of only low-level
            # objects and subgraphs. All invalid connections are pruned.
            self.reduce()

            # +~~ expansion
            # ensures that there is at most one signal connection at any inlet
            self.cascade_expansion()

            # fma replacement
            # convert [__mul~f ~f> __add~f] into [__fma~f]
            self.fma_replacement()

            # group all control receivers with the same name under one logical receiver
            self.group_control_receivers()

            # assign signal buffers to signal objects
            # Buffer assignment only takes place at the very end,
            # and does so recursively over all objects and subgraphs.
            # All objects are ordered before buffers are assigned.
            self.assign_signal_buffers()
        except HeavyException as e:
            e.notes = self.get_notices()
            e.notes["has_error"] = True
            e.notes["exception"] = e
            raise e

    def _remove_unused_inlet_connections(self):
        """ Remove connections from inlet object, if there are no incoming
            connections to it. This is a basic approach to pruning unused or
            unnecessary connections, which allow further optimisation later
            in the pipeline.
        """
        for i, cc in enumerate(self.inlet_connections):
            if len(cc) == 0:
                # make copy of outlet_connections list because it will be changed
                for c in list(self.inlet_objs[i].outlet_connections[0]):
                    self.disconnect_objects(c)

        # the recursive bit
        for o in [o for o in self.objs.values() if (o.type == "__graph")]:
            o._remove_unused_inlet_connections()

    def _resolved_outlet_type(self, outlet_index=0):
        # a graph's outlet type depends on the connections incident on the
        # corresponding outlet object
        connection_type_set = {c.type for c in self.outlet_objs[outlet_index].inlet_connections[0]}
        if len(connection_type_set) == 0:
            return "-->"  # object has no incident connections, default to -->
        elif len(connection_type_set) == 1:
            return list(connection_type_set)[0]
        else:
            raise HeavyException(f"{self} has multiple incident connections of differing type.\
                                  The outlet type cannot be explicitly resolved.")

    def _resolve_connection_types(self, obj_stack=None):
        """ Resolves the type of all connections before reduction to IR object types.
            If connections incident on an object are incompatible, they are either
            resolved, potentially by inserting conversion objects, or pruned.
            If a viable resolution cannot be found, an exception may be raised.
        """
        obj_stack = obj_stack or set()
        if self in obj_stack:
            return  # ensure that each object is only resolved once (no infinite loops)
        else:
            obj_stack.add(self)

        # start from all roots and resolve all connections downwards
        for o in [o for o in self.objs.values() if o.is_root()]:
            o._resolve_connection_types(obj_stack)

        # when all internal connections have been resolved,
        # continue the recursion and resolve all outgoing connections
        for c in [c for cc in self.outlet_connections for c in cc]:
            if c.is_mixed:
                # resolve connection type and update it
                connection_type = self._resolved_outlet_type(c.outlet_index)
                self.update_connection(c, [c.copy(type=connection_type)])
            c.to_object._resolve_connection_types(obj_stack)  # turtle on down

    def remap_send_receive(self):
        """ Recursively finds all signal send/receive objects and for each unique
            send name, get all of the sends, and all of the corresponding receives.
            Reconnect all incoming connections to a __add~f, and fan out the results
            to all of receivers' connections. At most one __add~f must
            be added, and a lot of connections remapped.
        """

        sends_dict = self.local_vars.get_registered_objects_for_type("send")
        receives_dict = self.local_vars.get_registered_objects_for_type("receive")

        for (name, send_objs) in sends_dict.items():

            # unconnected send/receives will be removed in the reduce step,
            # if necessary (depending on control or signal functionality).
            # All signal sends will also be removed, regardless of whether they
            # have corresponding receives or not. So here we really only need to
            # be concerned with connecting "normal" signal send/receives.

            # get all signal sends for a name
            s_list = [o for o in send_objs if o.has_inlet_connection_format("f")]
            # assume that all similarly named receives will also be signal format
            r_list = [o for o in receives_dict[name]]

            if len(s_list) == 0:
                continue  # there are no signals going into this send
            elif len(s_list) == 1:
                s = s_list[0]
                c = s.inlet_connections[0][0]  # the connection

                ir_var = HeavyIrObject("__var~f")
                ir_varset = HeavyIrObject("__varwrite~f", {"var_id": ir_var.id})
                s.graph.add_object(ir_var)
                s.graph.add_object(ir_varset)

                # move all connection to send object, to ir_varset
                for c in list(s.inlet_connections[0]):
                    s.graph.update_connection(c, [c.copy(to_object=ir_varset)])

                # move connections from receivers, to be from __var~f
                for o in r_list:
                    ir_varread = HeavyIrObject("__varread~f", {"var_id": ir_var.id})
                    o.graph.add_object(ir_varread)
                    for x in list(o.outlet_connections[0]):
                        o.graph.update_connection(x, [x.copy(from_object=ir_varread)])
            else:
                ir_vars = [HeavyIrObject("__var~f") for s in s_list]
                for i, o in enumerate(s_list):
                    ir_varset = HeavyIrObject("__varwrite~f", {"var_id": ir_vars[i].id})
                    o.graph.add_object(ir_vars[i])
                    o.graph.add_object(ir_varset)
                    for c in list(o.inlet_connections[0]):
                        o.graph.update_connection(c, [c.copy(to_object=ir_varset, inlet_index=0)])

                for r in r_list:
                    ir_add = HeavyIrObject("__add~f")
                    r.graph.add_object(ir_add)

                    for i, s in enumerate(s_list):
                        ir_varread = HeavyIrObject("__varread~f", {"var_id": ir_vars[i].id})
                        r.graph.add_object(ir_varread)
                        r.graph.connect_objects(Connection(
                            from_object=ir_varread,
                            outlet_index=0,
                            to_object=ir_add,
                            inlet_index=(0 if i == 0 else 1),
                            conn_type="~f>"
                        ))

                    for c in list(r.outlet_connections[0]):
                        r.graph.update_connection(c, [c.copy(from_object=ir_add)])

            # when all is said and done, remove the send and receive objects
            for o in s_list:
                o.graph.remove_object(o)
            for o in r_list:
                o.graph.remove_object(o)

    def reduce(self):
        """ Breaks this object into low-level objects. This method returns either
            the object that it is called on, or a graph. In case of a graph, it contains
            only low-level objects. Unnecessary connections are pruned. Because
            this method is called on graphs while they are being constructed, returned
            graphs contains no sub-graphs. Thus this method does not process subgraphs.
        """
        # refactor all constituent objects
        for (obj_id, o) in self.objs.copy().items():
            # use items and not iteritems so that object dictionary remains mutable

            # break the object into atomic (i.e. low-level) objects and
            # update connections. Replace the new representation with the old
            # one in the graph.
            objects, connections = o.reduce()

            # if x is the original object (the case with low-level objects),
            # then no change must be made
            if (len(objects) == 1) and (o in objects):
                continue

            # add the new objects
            for x in objects:
                self.add_object(x)

            # replace existing connections
            # control connection order is maintained
            for c in connections:
                # c is a tuple containing the old and new connections
                # the new connection is a possibly empty list of new connections,
                # indicating that there is no replacement
                self.update_connection(c[0], c[1])

            # remove the old object (and any remaining connections) from the graph
            self.remove_object(o, obj_id=obj_id)
            # o.id may not be the same as obj_id if the object comes from an abstraction

        # a graph is reduced in-place and does not change any connections
        return ({self}, [])

    def cascade_expansion(self):
        """ Turns implicit +~ into explicit cascading +~ trees.
            It is assumed that this simplification operation is run on a reduced graph.
        """
        for o in self.objs.copy().values():  # all items in the graph
            for i in range(len(o.inlet_connections)):  # for each inlet
                # if there is more than one signal connection
                # get all of the signal connections to an object at this inlet
                cc = [c for c in o.inlet_connections[i] if c.is_signal]
                if len(cc) > 1:
                    oL = HeavyIrObject("__add~f")
                    self.add_object(oL)

                    self.update_connection(
                        cc[0],
                        [Connection.copy(cc[0], to_object=oL, inlet_index=0)])
                    self.update_connection(
                        cc[1],
                        [Connection.copy(cc[1], to_object=oL, inlet_index=1)])

                    for j in range(2, len(cc)):
                        x = HeavyIrObject("__add~f")
                        self.add_object(x)

                        self.connect_objects(Connection(
                            from_object=oL,
                            outlet_index=0,
                            to_object=x,
                            inlet_index=0,
                            conn_type="~f>"))
                        self.update_connection(
                            cc[j],
                            [Connection.copy(cc[j], to_object=x, inlet_index=1)])

                        oL = x

                    # add a connection from the last +~ to this inlet
                    self.connect_objects(Connection(
                        from_object=oL,
                        outlet_index=0,
                        to_object=o,
                        inlet_index=i,
                        conn_type="~f>"))

            if o.type == "__graph":
                o.cascade_expansion()

    def fma_replacement(self):
        """ Replace:
            [__mul~f] ~f> [__add~f] with [__fma~f] or
            [__mul~f] ~f> [__sub~f] with [__fms~f]
        """
        # for all __mul~f objects
        for o in self.objs.copy().values():
            if o.type == "__mul~f" \
                    and len(o.inlet_connections[0]) == 1 \
                    and len(o.inlet_connections[1]) == 1 \
                    and len(o.outlet_connections[0]) == 1 \
                    and (o.outlet_connections[0][0].to_object.type == "__add~f"
                         or (o.outlet_connections[0][0].to_object.type == "__sub~f"
                             and o.outlet_connections[0][0].inlet_index == 0)) \
                    and len(o.outlet_connections[0][0].to_object.inlet_connections[0]) == 1 \
                    and len(o.outlet_connections[0][0].to_object.inlet_connections[1]) == 1 \
                    and len(o.outlet_connections[0][0].to_object.outlet_connections[0]) > 0:

                fma_type = "__fma~f" if o.outlet_connections[0][0].to_object.type == "__add~f" else "__fms~f"
                fma = HeavyIrObject(fma_type)
                self.add_object(fma)

                # move connection to left inlet of fma~
                c = o.inlet_connections[0][0]
                self.update_connection(c, [c.copy(to_object=fma)])

                # move connection to right inlet of fma~
                c = o.inlet_connections[1][0]
                self.update_connection(c, [c.copy(to_object=fma)])

                o_add = o.outlet_connections[0][0].to_object

                # move connection to third inlet of fma~ from +~ (not connected to *~)
                i = o.outlet_connections[0][0].inlet_index  # i is either 0 or 1
                # i ^ 1 # use the other inlet; 0 > 1, 1 > 0
                c = o_add.inlet_connections[i ^ 1][0]
                self.update_connection(c, [c.copy(to_object=fma, inlet_index=2)])

                # move all +~ outlet connections to fma~ outlet
                for c in o_add.outlet_connections[0]:
                    self.connect_objects(c.copy(from_object=fma))

                # remove old *~ and +~ objects from the graph
                # (along with any remaining connections)
                o.graph.remove_object(o)  # *~
                o_add.graph.remove_object(o_add)  # +~

            elif o.type == "__graph":
                o.fma_replacement()  # recurse through all subgraphs

    def group_control_receivers(self):
        """ Group all control receivers with the same name under one receiver
            with that name. This way only one message must be scheduled to hit
            all receivers.
        """

        for name, receivers in self.local_vars.get_registered_objects_for_type("__receive").items():

            # ensure that all receiver arguments are consistent
            extern = list(set([r.args["extern"] for r in receivers]) - set([None]))
            attributes = [r.args["attributes"] for r in receivers if len(r.args["attributes"]) > 0]
            scope = list(set([r.annotations["scope"] for r in receivers]))

            if len(extern) > 1:
                self.add_error("Parameter \"{0}\" has conflicting extern types: {1} != {2}".format(
                    name, extern[0], extern[1]))

            # NOTE(mhroth): not checking for conflicting scope types right now. Not sure what to do with it.

            if not all(attributes[0] == a for a in attributes):
                self.add_error("Conflicting min/max/default values for parameter \"{0}\"".format(name))

            # create a new receiver
            recv = HIrReceive("__receive",
                              args={
                                  "name": name,
                                  "extern": extern[0] if len(extern) > 0 else None,
                                  "attributes": attributes[0] if len(attributes) > 0 else {},
                              },
                              annotations={"scope": scope[0]})

            # add new receiver to the top level graph
            self.add_object(recv)

            # sort receivers by priority
            receivers.sort(key=lambda x: x.args["priority"], reverse=True)

            # move all connections to new receiver
            for r in receivers:
                for c in r.outlet_connections[0]:
                    self.connect_objects(c.copy(from_object=recv), require_intra_graph_connection=False)

            # remove old receivers
            for r in receivers:
                r.graph.remove_object(r)

    @property
    def does_process_signal(self):
        # this graph processes a signal if it contains any adc~ (__inlet
        # objects with index > 127)
        return any(o.does_process_signal for o in self.objs.values()) or \
            any(o.args["index"] > 127 for o in self.inlet_objs)

    def order_signal_objects(self):
        """ Places the signal objects in the correct order to be processed.
            Only the objects in this graph are ordered, stopping at the inlet objects.
        """
        # for all leaves of graph, get the parent objects and combine the lists
        self.signal_order = []  # ordered objects
        for o in [o for o in self.objs.values() if o.is_leaf()]:
            self.signal_order.extend(o.get_parent_order())

        # retain only objects that process a signal
        self.signal_order = [o for o in self.signal_order if o.does_process_signal]

    def assign_signal_buffers(self, buffer_pool=None):
        self.buffer_pool = buffer_pool or BufferPool()  # the top-level graph owns the buffer pool

        # before signal buffers can be assigned, the objects must be ordered
        self.order_signal_objects()

        for c_list in self.inlet_connections:
            c_list = [c for c in c_list if c.is_signal]  # only consider signal connections to inlet
            if len(c_list) == 0:
                continue  # no connections at this inlet, it already contains the zero buffer
            if len(c_list) == 1:
                c = c_list[0]  # get the connection

                # get the buffer at the outlet of the connected object
                buf = c.from_object.outlet_buffers[c.outlet_index]

                # assign the buffer to the outlet of the corresponding inlet object
                inlet_obj = self.inlet_objs[c.inlet_index]
                inlet_obj.outlet_buffers[0] = buf

                # ensure that the retain count is accurately reflected by the inlet's connections
                # some inlets are mixed with both signal and control connections, count only the signal connections
                self.buffer_pool.retain_buffer(buf,
                                               len([c for c in inlet_obj.outlet_connections[0] if c.is_signal]) - 1)
            else:
                raise HeavyException("Object {0} in graph {1} has {2} (> 1) signal inputs.".format(
                    inlet_obj,
                    inlet_obj.graph.file,
                    len(c_list)))

        # for all objects in the signal order
        for o in self.signal_order:
            o.assign_signal_buffers(self.buffer_pool)

        # assign the output buffers
        for i, outlet_obj in enumerate(self.outlet_objs):
            # only assign signal buffers to outlets that have incoming signal connections
            c_list = [c for c in outlet_obj.inlet_connections[0] if c.is_signal]
            if len(c_list) == 0:
                continue
            if len(c_list) == 1:
                c = c_list[0]  # get the connection
                buf = c.from_object.outlet_buffers[c.outlet_index]
                self.outlet_buffers[i] = buf
                self.buffer_pool.retain_buffer(buf, len(self.outlet_connections[i]) - 1)
            else:
                raise HeavyException("Object {0} in graph {1} has {2} (> 1) signal inputs.".format(
                    outlet_obj,
                    outlet_obj.graph.file,
                    len(c_list)))

    def __repr__(self):
        if self.xname is not None:
            # TODO(mhroth): does not handle nested subgraph
            return "__graph.{0}({1}/{2})".format(
                self.id,
                os.path.basename(self.file),
                self.xname)
        else:
            return "__graph.{0}({1})".format(self.id, os.path.basename(self.file))

    #
    # Intermediate Representation generators
    #

    def to_ir(self):
        """ Returns Heavy intermediate representation.
        """

        # the set of all input/output signal channels used by the graph
        input_channel_set = self.get_input_channel_set(recursive=True)
        output_channel_set = self.get_output_channel_set(recursive=True)

        return {
            "name": {
                "escaped": re.sub("\W", "_", self.xname),
                "display": self.xname
            },
            "objects": self.get_object_dict(),
            "init": {
                "order": self.get_ir_init_list()
            },
            "tables": self.get_ir_table_dict(),
            "control": {
                "receivers": self.get_ir_receiver_dict(),
                "sendMessage": self.get_ir_control_list()
            },
            "signal": {
                "numInputBuffers": max(input_channel_set) if len(input_channel_set) > 0 else 0,
                "numOutputBuffers": max(output_channel_set) if len(output_channel_set) > 0 else 0,
                "numTemporaryBuffers": {
                    "float": self.buffer_pool.num_buffers("~f>"),
                    "integer": self.buffer_pool.num_buffers("~i>")
                },
                "processOrder": self.get_ir_signal_list()
            }
        }

    def get_object_dict(self):
        # d = {o.id: o.get_object_dict() for o in self.objs.values() if o.type not in
        # ["inlet", "__inlet", "outlet", "__outlet"]}
        d = {}
        for o in self.objs.values():
            # NOTE(mhroth): a bit of a hack to remove unnecessary inlet and outlet ir objects
            if o.type not in ["inlet", "__inlet", "outlet", "__outlet"]:
                d.update(o.get_object_dict())
        return d

    def get_ir_init_list(self):
        """ Init list is returned with all signal objects at the front,
            in the order that they are processed. This is to reduce cache misses
            on the signal object state as the process function is executed.
        """
        init_list = [x for o in self.objs.values() for x in o.get_ir_init_list()]
        signal_list = self.get_ir_signal_list()
        s_init_list = [x["id"] for x in signal_list if x["id"] in init_list]
        i_init_list = [o_id for o_id in init_list if o_id not in s_init_list]
        ordered_init_list = s_init_list + i_init_list
        # ordered_init_list = list(OrderedDict.fromkeys(s_init_list + i_init_list))
        return ordered_init_list

    def get_ir_on_message(self, inlet_index):
        # pass the method through the inlet object, but only follow control connections
        x = []
        for c in self.inlet_objs[inlet_index].outlet_connections[0]:
            if c.is_control:
                x.extend(c.to_object.get_ir_on_message(c.inlet_index))
        return x

    def get_ir_table_dict(self):
        """ Returns a dictionary of all publicly visible tables at the root graph
            and their ids.
        """
        assert self.is_root_graph(), "This function should only be called from the root graph."
        d = self.local_vars.get_registered_objects_for_type("__table")
        # update(), because tables can be registered either as Heavy tables
        # or HeavyIr __tables. We need to be able to handle them both.
        d.update(self.local_vars.get_registered_objects_for_type("table"))

        e = {}
        for k, v in d.items():
            # escape table key to be used as the value for code stubs
            key = ("_" + k) if re.match("\d", k) else k
            if key not in e:
                e[key] = {
                    "id": v[0].id,
                    "display": k,
                    "hash": "0x{0:X}".format(HeavyLangObject.get_hash(k)),
                    "extern": v[0].args["extern"]
                }
        return e

    def get_ir_control_list(self):
        return [x for o in self.objs.values() for x in o.get_ir_control_list()]

    def get_ir_receiver_dict(self):
        # NOTE(mhroth): this code assumes that v is always an array of length 1,
        # as the grouping of control receivers should have grouped all same-named
        # receivers into one logical receiver.
        # NOTE(mhroth): a code-compatible name is only necessary for externed receivers
        return {(("_" + k) if re.match("\d", k) else k): {
            "display": k,
            "hash": "0x{0:X}".format(HeavyLangObject.get_hash(k)),
            "extern": v[0].args["extern"],
            "attributes": v[0].args["attributes"],
            "ids": [v[0].id]
        } for k, v in self.local_vars.get_registered_objects_for_type("__receive").items()}

    def get_ir_signal_list(self):
        return [x for o in self.signal_order for x in o.get_ir_signal_list()]



# Copyright 2015,2016 Enzien Audio, Ltd. All Rights Reserved.

from .PdObject import PdObject


class PdLetObject(PdObject):
    def __init__(self, obj_type, obj_args=None, pos_x=0, pos_y=0):
        assert obj_type in ["inlet", "inlet~", "outlet", "outlet~"]
        PdObject.__init__(self, obj_type, obj_args, pos_x, pos_y)
        self.let_index = 0

    def get_outlet_connection_type(self, outlet_index=0):
        if len(self.obj_args) > 0 and self.obj_args[0] in ["-->", "~f>", "~i>", "-~>"]:
            return self.obj_args[0]
        else:
            return PdObject.get_outlet_connection_type(self)

    def to_hv(self):
        return {
            "type": self.obj_type.strip("~"),
            "args": {
                "name": "",  # Pd does not give an inlet name
                "index": self.let_index,
                "type": self.get_outlet_connection_type()
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

import random
import string


class MaxObject:
    def __init__(self, obj_type, obj_args=None, obj_id=None, pos_x=0, pos_y=0):
        self.obj_type = obj_type
        self.obj_args = obj_args or []
        self.obj_id = obj_id or "".join(random.choice(string.ascii_letters + string.digits) for _ in range(5))
        self.pos_x = int(pos_x)
        self.pos_y = int(pos_y)

    def get_outlet_connection_type(self, outlet_index=0):
        """Returns the outlet connection type of this Max object.
        For the sake of convenience, the connection type is reported in
        Heavy's format.
        """
        return "~f>" if self.obj_type.endswith("~") else "-->"

    @classmethod
    def get_supported_objects(clazz):
        """Returns a list of Max objects that this class can parse.
        """
        # TODO(mhroth): return [] ?
        raise NotImplementedError()

    def to_hv(self):
        """Returns the HeavyLang JSON representation of this object.
        """
        raise NotImplementedError()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import jinja2
import os
import shutil
import time
from ..buildjson import buildjson


class c2bela:
    """ Generates a makefile to compile source to object files suitable for the
        bela platform.
    """

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs,
                patch_name=None, num_input_channels=0, num_output_channels=0,
                copyright=None, verbose=False):

        tick = time.time()
        patch_name = patch_name or "heavy"

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over generated C source files
            shutil.copytree(c_src_dir, os.path.join(out_dir, "source"))

            # initialise the jinja template environment
            env = jinja2.Environment()
            env.loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))

            # make sure directories exist
            makefile_path = os.path.join(out_dir, "linux", "Makefile")
            if not os.path.exists(os.path.dirname(makefile_path)):
                os.makedirs(os.path.dirname(makefile_path))

            # generate linux makefile
            with open(makefile_path, "w") as f:
                f.write(env.get_template("linux/Makefile").render(name=patch_name))

            # generate build json file
            buildjson.generate_json(
                out_dir,
                linux_armv7a_args=["-j"])

            return {
                "stage": "c2bela",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2bela",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject


class HLangLine(HeavyLangObject):
    """ Translates HeavyLang object [line] to HeavyIR [__line] or [__line~f].
    """

    def __init__(self, obj_type, args, graph, annotations=None):
        assert obj_type == "line"
        HeavyLangObject.__init__(self, "line", args, graph,
                                 num_inlets=2,
                                 num_outlets=1,
                                 annotations=annotations)

    def reduce(self):
        if self.has_outlet_connection_format("f"):
            x = HeavyIrObject("__line~f", self.args)
        elif self.has_outlet_connection_format("c"):
            x = HeavyIrObject("__line", self.args)
        else:
            self.add_error("Unknown inlet configuration.")

        return ({x}, self.get_connection_move_list(x))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import random

from .HeavyLangObject import HeavyLangObject
from .HeavyParser import HeavyParser


class HLangNoise(HeavyLangObject):
    """ Handles the HeavyLang "noise" object.
    """

    def __init__(self, obj_type, args, graph, annotations=None):
        assert obj_type == "noise"
        HeavyLangObject.__init__(self, "noise", args, graph,
                                 num_inlets=1,
                                 num_outlets=1,
                                 annotations=annotations)

    def reduce(self):
        seed = int(random.uniform(1, 2147483647))  # assign a random 32-bit seed
        noise_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "./hvlib/noise.hv.json")
        x = HeavyParser.graph_from_file(noise_path, graph_args={"seed": seed})
        x.reduce()
        # TODO(mhroth): deal with control input
        return ({x}, self.get_connection_move_list(x))



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

from .MaxObject import MaxObject


class MaxOutletObject(MaxObject):
    def __init__(self, outlet_type=None, obj_id=None, pos_x=0, pos_y=0):
        MaxObject.__init__(self, "outlet", None, obj_id, pos_x, pos_y)
        self.__outlet_type = "~f>" if outlet_type == "signal" else "-->"

    def set_index(self, index):
        """Sets the index order of this outlet.
        """
        self.__index = index

    def get_outlet_connection_type(self, outlet_index=0):
        return self.__outlet_type

    def to_hv(self):
        return {
            "type": "outlet",
            "args": {
                "name": "",  # max does not give an outlet name
                "index": self.__index,
                "type": self.__outlet_type
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject


class HLangVario(HeavyLangObject):

    def __init__(self, obj_type, args, graph, annotations=None):
        HeavyLangObject.__init__(self, "vario", args, graph, annotations=annotations)

    def reduce(self):
        var_obj = self.graph.resolve_object_for_name(self.name, "var")
        if var_obj is None:
            raise HeavyException("No corresponding \"var\" object found for object {0} in file {1}".format(
                self,
                self.graph.file))

        if self.has_inlet_connection_format("f") and self.has_outlet_connection_format("_"):
            x = HeavyIrObject("__varwrite~f", {"var_id": var_obj.id})
            return {x}, self.get_connection_move_list(x)

        elif self.has_inlet_connection_format("i") and self.has_outlet_connection_format("_"):
            x = HeavyIrObject("__varwrite~i", {"var_id": var_obj.id})
            return {x}, self.get_connection_move_list(x)

        elif self.has_inlet_connection_format("_") and self.has_outlet_connection_format("f"):
            x = HeavyIrObject("__varread~f", {"var_id": var_obj.id})
            return {x}, self.get_connection_move_list(x)

        elif self.has_inlet_connection_format("_") and self.has_outlet_connection_format("i"):
            x = HeavyIrObject("__varread~i", {"var_id": var_obj.id})
            return {x}, self.get_connection_move_list(x)

        else:
            raise Exception()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject
from .pdowl import parse_pd_owl_args, PdOwlException


class PdSendObject(PdObject):
    def __init__(self, obj_type, obj_args=None, pos_x=0, pos_y=0):
        assert obj_type in ["s", "send", "s~", "send~", "throw~"]
        PdObject.__init__(self, obj_type, obj_args, pos_x, pos_y)

        self.__send_name = ""
        self.__extern_type = None
        self.__attributes = {}

        try:
            # send objects don't necessarily need to have a name
            self.__send_name = self.obj_args[0]

            # only extern control rate sends
            if obj_type in ["s", "send"]:
                if self.obj_args[1] == "@hv_param":
                    self.__extern_type = "param"
                elif self.obj_args[1] == "@hv_event":
                    self.__extern_type = "event"
        except Exception:
            pass

        if '@owl' in self.obj_args or '@owl_param' in self.obj_args:
            try:
                pd_owl_args = parse_pd_owl_args(self.obj_args)
                self.__attributes.update(pd_owl_args)
                self.__extern_type = "param"  # make sure output code is generated
            except PdOwlException as e:
                self.add_error(e)

    def validate_configuration(self):
        if len(self.obj_args) == 0:
            self.add_warning(
                "No name was given to this {0} object. "
                "It should have a name to reduce the risk of errors.".format(self.obj_type),
                NotificationEnum.WARNING_USELESS_OBJECT)
        if len(self._inlet_connections.get("0", [])) == 0:
            self.add_warning(
                "This object has no inlet connections. "
                "It does nothing and will be removed.",
                NotificationEnum.WARNING_USELESS_OBJECT)
        if self.obj_type in ["s", "send"] and len(self._inlet_connections.get("1", [])) > 0:
            self.add_error(
                "Connections to the right inlet of a send object "
                "are not supported. A name should be given.",
                NotificationEnum.ERROR_MISSING_REQUIRED_ARGUMENT)

    def to_hv(self):
        # note: control rate send/receive objects should not modify their name argument
        names = {
            "s": "",
            "send": "",
            "s~": "sndrcv_sig_",
            "send~": "sndrcv_sig_",
            "throw~": "thrwctch_sig_"
        }

        return {
            "type": "send",
            "args": {
                "name": names[self.obj_type] + self.__send_name,
                "extern": self.__extern_type,
                "attributes": self.__attributes,
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            },
            "annotations": {
                "scope": "public"
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict

from .HeavyException import HeavyException


class BufferPool:

    def __init__(self):
        # the idea is that the same buffer is reused as quickly as possible so that it doesn't need
        # to be moved around in the cache. It does not give substantially different results
        # from a Counter-based implementation, but it is more consistent and predictable.
        self.pool = {
            "~f>": defaultdict(list),
            "~i>": defaultdict(list)
        }

    def num_buffers(self, connection_type=None):
        """ Returns the number of buffers with the given retain count. By default returns
            the size of the entire pool. Number of buffers per connection type can also be retrieved.
        """
        if connection_type is None:
            return self.num_buffers("~f>") + self.num_buffers("~i>")
        elif connection_type in self.pool:
            return sum(len(v) for v in self.pool[connection_type].values())
        else:
            raise HeavyException("Unknown connection type: \"{0}\"".format(connection_type))

    def get_buffer(self, connection_type, count=1, excludeSet=None):
        """ Returns a currently unused buffer. The buffer can be assigned a retain count. An optional
            exclude set can also be supplied, ensuring that the returned buffer is not one of them.
        """
        excludeSet = excludeSet if excludeSet is not None else set()

        pool = self.pool[connection_type]

        # get the most recently used, unused buffer
        b = next((b for b in reversed(pool[0]) if b not in excludeSet), None)
        if b is not None:
            pool[0].remove(b)
        else:
            # if we get here, then no available buffer was found. Create a new one.
            b = (connection_type, self.num_buffers(connection_type))  # new buffer index for the given type
        pool[count].append(b)
        return b

    def retain_buffer(self, b, count=1):
        """ Increases the retain count of the buffer. Returns the new count.
        """
        # adc~ and ZERO_BUFFER are special. They cannot be retained.
        if b[0] in ["zero", "input"]:
            return 0
        else:
            pool = self.pool[b[0]]
            for k, v in pool.items():
                if b in v:
                    v.remove(b)
                    pool[k + count].append(b)
                    return k + count  # return the new retain count
            raise HeavyException("{0} not found in BufferPool!".format(b))

    def release_buffer(self, b, count=1):
        """ Reduces the retain count of the buffer. Returns the new count.
        """
        # adc~, ZERO_BUFFER, send~ buffers are special. They can not be released.
        # if the buffer is otherwise unknown (as may be in the case that objects provide their own),
        # they cannot be released
        if b[0] in ["zero", "input"]:
            return 0
        else:
            pool = self.pool[b[0]]
            for k, v in pool.items():
                if b in v:
                    v.remove(b)
                    pool[k - count].append(b)
                    return k - count  # return the new retain count
            raise HeavyException("{0} not found in BufferPool!".format(b))

    def __repr__(self):
        return self.pool["~f>"].__repr__() + self.pool["~i>"].__repr__()



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

import json
import os
import re

from .MaxObject import MaxObject


class HeavyObject(MaxObject):

    with open(os.path.join(os.path.dirname(__file__), "../../core/json/heavy.lang.json"), "r") as f:
        __HEAVY_LANG_OBJS = json.loads(f.read())

    with open(os.path.join(os.path.dirname(__file__), "../../core/json/heavy.ir.json"), "r") as f:
        __HEAVY_IR_OBJS = json.loads(f.read())

    __re_dollar = re.compile("\$(\d+)")

    def __init__(self, obj_type, obj_args=None, obj_id=None, pos_x=0, pos_y=0):
        MaxObject.__init__(self, obj_type, obj_args, obj_id, pos_x, pos_y)

        # get the object dictionary (note that it is NOT a copy)
        if self.is_hvlang:
            self.__obj_dict = HeavyObject.__HEAVY_LANG_OBJS[obj_type]
        elif self.is_hvir:
            self.__obj_dict = HeavyObject.__HEAVY_IR_OBJS[obj_type]
        else:
            raise Exception("{0} is not a Heavy Lang or IR object.".format(obj_type))

        # resolve arguments
        obj_args = obj_args or []
        self.obj_args = {}
        for i, a in enumerate(self.__obj_dict["args"]):
            # if the argument exists and is not an unresolved dollar argument
            if len(obj_args) > i and not HeavyObject.__re_dollar.findall(obj_args[i]):
                # TODO(mhroth): need to force type of argument?
                self.obj_args[a["name"]] = obj_args[i]
            else:
                # the default argument is required
                if a["required"]:
                    raise Exception("Required argument \"{0}\" to object {1} not present: {2}".format(
                        a["name"],
                        obj_type,
                        obj_args))
                else:
                    self.obj_args[a["name"]] = a["default"]

    @classmethod
    def is_heavy(clazz, obj_type):
        return (obj_type in HeavyObject.__HEAVY_LANG_OBJS) or \
            (obj_type in HeavyObject.__HEAVY_IR_OBJS)

    @property
    def is_hvlang(self):
        return self.obj_type in HeavyObject.__HEAVY_LANG_OBJS

    @property
    def is_hvir(self):
        return self.obj_type in HeavyObject.__HEAVY_IR_OBJS

    def get_outlet_connection_type(self, outlet_index):
        # TODO(mhroth): it's stupid that hvlang and hvir json have different
        # data formats here
        if self.is_hvlang:
            return self.__obj_dict["outlets"][outlet_index]["connectionType"]
        elif self.is_hvir:
            return self.__obj_dict["outlets"][outlet_index]
        else:
            raise Exception()

    def to_hv(self):
        return {
            "type": self.obj_type,
            "args": self.obj_args,
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

from .MaxObject import MaxObject


class MaxBinopObject(MaxObject):
    # a translation dictionary from max object to corresponding heavy object
    __MAX_HEAVY_DICT = {
        "+": "+",
        "+~": "+",
        "-": "-",
        "-~": "-",
        "*": "*",
        "*~": "*"
    }

    def __init__(self, obj_type, obj_args=None, obj_id=None, pos_x=0, pos_y=0):
        assert MaxBinopObject.is_binop(obj_type)
        MaxObject.__init__(self, obj_type, obj_args, obj_id, pos_x, pos_y)

    @classmethod
    def is_binop(clazz, obj_type):
        return obj_type in MaxBinopObject.__MAX_HEAVY_DICT

    @classmethod
    def get_supported_objects(clazz):
        return MaxBinopObject.__MAX_HEAVY_DICT.keys()

    def to_hv(self):
        return {
            "type": MaxBinopObject.__MAX_HEAVY_DICT[self.obj_type],
            "args": {
                "k": float(self.obj_args[0]) if len(self.obj_args) > 0 else 0.0
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

import os

from .Connection import Connection
from .MaxObject import MaxObject


class MaxGraph(MaxObject):
    def __init__(self, obj_type=None, obj_args=None, obj_id=None, pos_x=0, pos_y=0, max_path=None):
        MaxObject.__init__(self, "patcher", obj_args, obj_id, pos_x, pos_y)

        # file location of this graph
        self.__max_path = max_path

        self.__objs = {}
        self.__connections = []

        self.__inlet_objects = []
        self.__outlet_objects = []

    def add_object(self, obj, obj_id=None):
        obj_id = obj_id or obj.obj_id
        assert obj_id not in self.__objs
        self.__objs[obj_id] = obj

        if obj.obj_type == "inlet":
            obj.set_index(len(self.__inlet_objects))
            self.__inlet_objects.append(obj)
        elif obj.obj_type == "outlet":
            obj.set_index(len(self.__outlet_objects))
            self.__outlet_objects.append(obj)

    def add_connection(self, from_id, from_outlet, to_id, to_inlet):
        assert from_id in self.__objs and to_id in self.__objs
        c = Connection(
            from_id, from_outlet,
            to_id, to_inlet,
            self.__objs[from_id].get_outlet_connection_type(from_outlet))
        self.__connections.append(c)

    def get_outlet_connection_type(self, outlet_index):
        return self.__outlet_objects[outlet_index].get_outlet_connection_type()

    def to_hv(self):
        return {
            "type": "graph",
            "imports": [],
            "args": [],
            "objects": {k: v.to_hv() for k, v in self.__objs.iteritems()},
            "connections": [c.to_hv() for c in self.__connections],
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }

    def __repr__(self):
        return "graph.{0}({1})".format(
            self.obj_id,
            os.path.basename(self.__max_path))



# import datetime
import hashlib
import os
import shutil
import time
import jinja2
from ..buildjson import buildjson
from ..copyright import copyright_manager


class c2dpf:
    """ Generates a DPF wrapper for a given patch.
    """

    @classmethod
    def filter_uniqueid(clazz, s):
        """ Return a unique id (in hexadcemial) for the VST interface.
        """
        s = hashlib.md5(s.encode('utf-8'))
        s = s.hexdigest().upper()[0:8]
        s = f"0x{s}"
        return s

    @classmethod
    def compile(clazz, c_src_dir, out_dir, externs,
                patch_name=None, num_input_channels=0, num_output_channels=0,
                copyright=None, verbose=False):

        tick = time.time()

        receiver_list = externs['parameters']['in']

        patch_name = patch_name or "heavy"

        copyright_c = copyright_manager.get_copyright_for_c(copyright)
        # copyright_plist = copyright or u"Copyright {0} Enzien Audio, Ltd." \
        #     " All Rights Reserved.".format(datetime.datetime.now().year)

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over static files
            shutil.copytree(os.path.join(os.path.dirname(__file__), "static"), out_dir)

            # copy over generated C source files
            source_dir = os.path.join(out_dir, "source")
            shutil.copytree(c_src_dir, source_dir)

            # initialize the jinja template environment
            env = jinja2.Environment()
            env.filters["uniqueid"] = c2dpf.filter_uniqueid

            env.loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))

            # generate DPF wrapper from template
            dpf_h_path = os.path.join(source_dir, f"HeavyDPF_{patch_name}.hpp")
            with open(dpf_h_path, "w") as f:
                f.write(env.get_template("HeavyDPF.hpp").render(
                    name=patch_name,
                    class_name=f"HeavyDPF_{patch_name}",
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    receivers=receiver_list,
                    copyright=copyright_c))
            dpf_cpp_path = os.path.join(source_dir, f"HeavyDPF_{patch_name}.cpp")
            with open(dpf_cpp_path, "w") as f:
                f.write(env.get_template("HeavyDPF.cpp").render(
                    name=patch_name,
                    class_name=f"HeavyDPF_{patch_name}",
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    receivers=receiver_list,
                    pool_sizes_kb=externs["memoryPoolSizesKb"],
                    copyright=copyright_c))
            dpf_h_path = os.path.join(source_dir, "DistrhoPluginInfo.h")
            with open(dpf_h_path, "w") as f:
                f.write(env.get_template("DistrhoPluginInfo.h").render(
                    name=patch_name,
                    class_name=f"HeavyDPF_{patch_name}",
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    receivers=receiver_list,
                    pool_sizes_kb=externs["memoryPoolSizesKb"],
                    copyright=copyright_c))

            # generate list of Heavy source files
            # files = os.listdir(source_dir)

            # ======================================================================================
            # Linux
            #
            # linux_path = os.path.join(out_dir, "linux")
            # os.makedirs(linux_path)

            with open(os.path.join(source_dir, "Makefile"), "w") as f:
                f.write(env.get_template("Makefile").render(
                    name=patch_name,
                    class_name="HeavyDPF_" + patch_name))

            buildjson.generate_json(
                out_dir,
                linux_x64_args=["-j"])
            # macos_x64_args=["-project", "{0}.xcodeproj".format(patch_name), "-arch",
            #                 "x86_64", "-alltargets"],
            # win_x64_args=["/property:Configuration=Release", "/property:Platform=x64",
            #               "/t:Rebuild", "{0}.sln".format(patch_name), "/m"],
            # win_x86_args=["/property:Configuration=Release", "/property:Platform=x86",
            #               "/t:Rebuild", "{0}.sln".format(patch_name), "/m"])

            return {
                "stage": "c2dpf",
                "notifs": {
                    "has_error": False,
                    "exception": None,
                    "warnings": [],
                    "errors": []
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": os.path.basename(dpf_h_path),
                "compile_time": time.time() - tick
            }

        except Exception as e:
            return {
                "stage": "c2dpf",
                "notifs": {
                    "has_error": True,
                    "exception": e,
                    "warnings": [],
                    "errors": [{
                        "enum": -1,
                        "message": str(e)
                    }]
                },
                "in_dir": c_src_dir,
                "in_file": "",
                "out_dir": out_dir,
                "out_file": "",
                "compile_time": time.time() - tick
            }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from .HeavyIrObject import HeavyIrObject


class HIrSend(HeavyIrObject):
    """ A specific implementation of the __send object.
    """

    def __init__(self, obj_type, args=None, graph=None, annotations=None):
        HeavyIrObject.__init__(self, "__send", args=args, graph=graph, annotations=annotations)
        if args["extern"]:
            # output parameters must contain only alphanumeric characters or underscores,
            # so that the names can be easily and transparently turned into code
            if re.search("\W", args["name"]):
                self.add_error(f"Parameter and Event names may only contain \
                                alphanumeric characters or underscore: '{args['name']}'")

    def get_ir_control_list(self):
        receive_objs = self.graph.resolve_objects_for_name(self.name, "__receive")
        on_message_list = [x for o in receive_objs for x in o.get_ir_on_message(inlet_index=0)]
        return [{
            "id": self.id,
            "onMessage": [on_message_list],
            "extern": self.args["extern"],
            "hash": self.args["hash"],
            "display": self.args["name"],
            "name": (("_" + self.args["name"]) if re.match("\d", self.args["name"]) else self.args["name"])
        }]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject


class HLangDac(HeavyLangObject):
    """ Translates HeavyLang [dac] to HeavyIR [__add~f].
    """

    def __init__(self, obj_type, args, graph, annotations=None):
        assert obj_type == "dac"
        HeavyLangObject.__init__(
            self,
            obj_type,
            args, graph,
            num_inlets=len(args[HeavyLangObject._HEAVY_LANG_DICT["dac"]["args"][0]["name"]]),
            num_outlets=0,
            annotations=annotations)

    def reduce(self):
        objects = set()
        connections = []

        # reduce a HeavyLang dac to a number of individual HeavyIR __add~f objects
        for i, channel_index in enumerate(self.args["channels"]):
            if len(self.inlet_connections[i]) > 0:  # if there are any connections to this inlet
                x = HeavyIrObject("__add~f")
                x.inlet_buffers[1] = ("output", channel_index - 1)  # channel indicies are one-indexed
                x.outlet_buffers[0] = ("output", channel_index - 1)
                objects.add(x)

                for c in self.inlet_connections[i]:
                    connections.append((c, [c.copy(to_object=x, inlet_index=0)]))

        return (objects, connections)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import Counter
from collections import OrderedDict
import decimal
import os
import re

from .HeavyObject import HeavyObject
from .HeavyGraph import HeavyGraph              # pre-converted Heavy graphs
from .HvSwitchcase import HvSwitchcase          # __switchcase
from .PdAudioIoObject import PdAudioIoObject    # adc~/dac~
from .PdBinopObject import PdBinopObject        # binary arithmatic operators
from .PdGraph import PdGraph                    # canvas
from .PdLetObject import PdLetObject            # inlet/inlet~/outlet/outlet~
from .PdMessageObject import PdMessageObject    # msg
from .PdPackObject import PdPackObject          # pack
from .PdReceiveObject import PdReceiveObject    # r/r~/receive/receive~/catch~
from .PdRouteObject import PdRouteObject        # route
from .PdSelectObject import PdSelectObject      # select/sel
from .PdSendObject import PdSendObject          # s/s~/send/send~/throw~
from .PdTriggerObject import PdTriggerObject    # trigger/t
from .PdTableObject import PdTableObject        # table
from .PdUnpackObject import PdUnpackObject      # unpack
from .PdLibSignalGraph import PdLibSignalGraph  # pd/lib abstraction connection checks

from .NotificationEnum import NotificationEnum


class PdParser:

    # library search paths
    __HVLIB_DIR = os.path.join(os.path.dirname(__file__), "libs", "heavy")
    __HVLIB_CONVERTED_DIR = os.path.join(os.path.dirname(__file__), "libs", "heavy_converted")
    __PDLIB_DIR = os.path.join(os.path.dirname(__file__), "libs", "pd")
    __PDLIB_CONVERTED_DIR = os.path.join(os.path.dirname(__file__), "libs", "pd_converted")

    # detect a dollar argument in a string
    __RE_DOLLAR = re.compile("\$(\d+)")

    # detect width parameter e.g. "#X obj 172 79 t b b, f 22;"
    __RE_WIDTH = re.compile(", f \d+$")

    def __init__(self):
        # the current global value of $0
        # Note(joe): set a high starting value to avoid potential user naming conflicts
        self.__DOLLAR_ZERO = 1000

        # a counter of all Pd objects in the graph
        self.obj_counter = Counter()

        # search paths at this graph level
        self.__search_paths = []

    @classmethod
    def get_supported_objects(clazz):
        """ Returns a set of all pd objects names supported by the parser.
        """
        pd_objects = [os.path.splitext(f)[0] for f in os.listdir(clazz.__PDLIB_DIR) if f.endswith(".pd")]
        pd_objects.extend(PdParser.__PD_CLASSES.keys())
        return pd_objects

    @classmethod
    def __get_hv_args(clazz, pd_path):
        """ Pre-parse the file for Heavy arguments, such that they are available
            as soon as a graph is created.
        """
        num_canvas = -1
        hv_arg_dict = OrderedDict()
        hv_arg_list = None
        with open(pd_path, "r") as f:
            for li in f:
                if li.startswith("#N canvas"):
                    hv_arg_list = []
                    hv_arg_dict[li.rstrip(";\r\n")] = hv_arg_list
                    num_canvas += 1
                elif "@hv_arg" in li:
                    hv_arg_list.append(li.rstrip(";\r\n"))
                elif li.startswith("#X restore"):
                    num_canvas -= 1
                    hv_arg_list = list(hv_arg_dict.values())[num_canvas]
        return hv_arg_dict

    @classmethod
    def __get_pd_line(clazz, pd_path):
        concat = ""  # concatination state
        with open(pd_path, "r") as f:
            for li in f:
                # concatenate split lines in the Pd file here
                li = li.rstrip("\r\n")  # account for windows CRLF
                if li.endswith(";") and not li.endswith("\;"):
                    out = li[:-1]  # remove single ";"
                    if len(concat) > 0:
                        out = concat + " " + out
                        concat = ""  # reset concatenation state
                    yield out
                else:
                    concat = (concat + " " + li) if len(concat) > 0 else li

    def add_absolute_search_directory(self, search_dir):
        if os.path.isdir(search_dir):
            self.__search_paths.append(search_dir)
            return True
        else:
            return False

    def add_relative_search_directory(self, search_dir):
        search_dir = os.path.abspath(os.path.join(
            self.__search_paths[0],
            search_dir))
        return self.add_absolute_search_directory(search_dir)

    def find_abstraction_path(self, local_dir, abs_name):
        """ Finds the full path for an abstraction.
            Checks the local directory first, then all declared paths.
        """

        abs_filename = abs_name + ".pd"

        # check local directory first
        abs_path = os.path.join(os.path.abspath(local_dir), abs_filename)
        if os.path.isfile(abs_path):
            return abs_path

        # check search paths in reverse order (last added search path first)
        for d in reversed(self.__search_paths):
            abs_path = os.path.join(d, abs_filename)
            if os.path.isfile(abs_path):
                return abs_path

        return None

    def graph_from_file(self, file_path, obj_args=None, pos_x=0, pos_y=0, is_root=True, pd_graph_class=PdGraph):
        """ Instantiate a PdGraph from a file.
            Note that obj_args does not include $0.
            @param pd_graph_class  The python class to handle specific graph types
        """
        # add main patch directory. The first entry of self.__search_paths is
        # assumed to be the root path of the whole system

        if is_root:
            self.__search_paths.append(os.path.dirname(file_path))

        file_hv_arg_dict = PdParser.__get_hv_args(file_path)
        file_iterator = PdParser.__get_pd_line(file_path)
        canvas_line = file_iterator.__next__()

        self.__DOLLAR_ZERO += 1  # increment $0
        graph_args = [self.__DOLLAR_ZERO] + (obj_args or [])

        if not canvas_line.startswith("#N canvas"):
            g = pd_graph_class(graph_args, file_path, pos_x, pos_y)
            g.add_error("Pd files must begin with \"#N canvas\": {0}".format(canvas_line))
            return g

        g = self.graph_from_canvas(
            file_iterator,
            file_hv_arg_dict,
            canvas_line,
            graph_args,
            file_path,
            pos_x, pos_y,
            is_root,
            pd_graph_class)

        if is_root:
            if g.get_notices()["errors"]:
                # return the graph early here as there are already errors and it is
                # clearly invalid, avoids triggering unrelated errors in validation
                return g
            g.validate_configuration()

        return g

    def graph_from_canvas(self, file_iterator, file_hv_arg_dict, canvas_line, graph_args,
                          pd_path, pos_x=0, pos_y=0, is_root=False, pd_graph_class=PdGraph):
        """ Instantiate a PdGraph from an existing canvas.
            Note that graph_args includes $0.
            @param file_hv_arg_dict  A dictionary containing all Heavy argument lines
            for each "#N canvas" in this file.
            @param canvas_line  The "#N canvas" which initiates this canvas.
            @param pd_graph_class  The python class to handle specific graph types
        """
        obj_array = None  # an #A (table) object which is currently being parsed

        g = pd_graph_class(graph_args, pd_path, pos_x, pos_y)

        # parse and add all Heavy arguments to the graph
        for li in file_hv_arg_dict[canvas_line]:
            line = li.split()
            assert line[4] == "@hv_arg"
            is_required = (line[9] == "true")
            default_value = HeavyObject.force_arg_type(line[8], line[7]) \
                if not is_required else None
            g.add_hv_arg(
                arg_index=int(line[5][2:]) - 1,  # strip off the leading "\$" and make index zero-based
                name=line[6],
                value_type=line[7],
                default_value=default_value,
                required=is_required)

        try:  # this try will capture any critical errors
            for li in file_iterator:
                # remove width parameter
                line = PdParser.__RE_WIDTH.sub("", li).split()

                if line[0] == "#N":
                    if line[1] == "canvas":
                        x = self.graph_from_canvas(
                            file_iterator,
                            file_hv_arg_dict,
                            canvas_line=li,
                            graph_args=graph_args,  # subpatch inherits parent graph arguments, including $0
                            pd_path=pd_path,
                            pos_x=int(line[2]),
                            pos_y=int(line[3]))
                        g.add_object(x)

                elif line[0] == "#X":
                    if line[1] == "restore":
                        if len(line) > 5 and line[5] == "@hv_obj":
                            obj_args = PdParser.__resolve_object_args(
                                obj_type=line[6],
                                obj_args=line[7:],
                                graph=g,
                                raise_on_failure=False,
                                is_root=is_root)
                            if line[6] == "__switchcase":
                                x = HvSwitchcase(
                                    obj_type=line[6],
                                    obj_args=obj_args,
                                    pos_x=int(line[2]),
                                    pos_y=int(line[3]))
                            else:
                                x = HeavyObject(
                                    obj_type=line[6],
                                    obj_args=obj_args,
                                    pos_x=int(line[2]),
                                    pos_y=int(line[3]))
                            return x  # return a Heavy object instead of a graph
                        else:
                            # are we restoring an array object?
                            # do some final sanity checks
                            if obj_array is not None:
                                declared_size = obj_array.obj_args["size"]
                                values_size = len(obj_array.obj_args["values"])
                                if declared_size != values_size:
                                    new_size = max(declared_size, values_size)
                                    obj_array.add_warning(
                                        "Table \"{0}\" was declared as having {1} values, "
                                        "but {2} were supplied. It will be resized to {3} "
                                        "values (any unsupplied values will be zeroed).".format(
                                            obj_array.obj_args["name"],
                                            declared_size,
                                            values_size,
                                            new_size))
                                    obj_array.obj_args["size"] = new_size
                                    if new_size < declared_size:
                                        obj_array.obj_args["values"] = obj_args["values"][:new_size]
                                    else:
                                        obj_array.obj_args["values"].extend([0.0 for _ in
                                                                             range(new_size - declared_size)])
                                obj_array = None  # done parsing the array

                            # set the subpatch name
                            g.subpatch_name = " ".join(line[5:]) if len(line) > 5 else "subpatch"
                            return g  # pop the graph

                    elif line[1] == "text":
                        # @hv_arg arguments are pre-parsed
                        # TODO(mhroth): is it necessary to split the entire line at once?
                        # always add the comment to the graph, regardless
                        self.obj_counter["text"] += 1
                        g.add_object(HeavyObject(
                            obj_type="comment",
                            obj_args=[" ".join(line[4:])],
                            pos_x=int(line[2]),
                            pos_y=int(line[3])))

                    elif line[1] == "obj":
                        x = None  # a PdObject
                        if len(line) > 4:
                            obj_type = line[4]
                            # sometimes objects have $ arguments in them as well
                            obj_type = PdParser.__resolve_object_args(
                                obj_type=obj_type,
                                obj_args=[obj_type],
                                graph=g,
                                is_root=is_root)[0]
                            obj_args = PdParser.__resolve_object_args(
                                obj_type=obj_type,
                                obj_args=line[5:],
                                graph=g,
                                is_root=is_root)
                        else:
                            g.add_warning(
                                "This graph contains an empty object. "
                                "It should be removed or defined.",
                                NotificationEnum.WARNING_EMPTY_OBJECT)
                            g.add_object(HeavyObject(
                                obj_type="comment",
                                obj_args=["null object placeholder"],
                                pos_x=int(line[2]),
                                pos_y=int(line[3])))
                            continue

                        # do we have an abstraction for this object?
                        abs_path = self.find_abstraction_path(os.path.dirname(pd_path), obj_type)
                        if abs_path is not None and not g.is_abstraction_on_call_stack(abs_path):
                            # ensure that infinite recursion into abstractions is not possible
                            x = self.graph_from_file(
                                file_path=abs_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False)

                        # is this object in lib/pd_converted?
                        elif os.path.isfile(os.path.join(PdParser.__PDLIB_CONVERTED_DIR, obj_type + ".hv.json")):
                            self.obj_counter[obj_type] += 1
                            hv_path = os.path.join(PdParser.__PDLIB_CONVERTED_DIR, obj_type + ".hv.json")
                            x = HeavyGraph(
                                hv_path=hv_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        # is this object in lib/heavy_converted?
                        elif os.path.isfile(os.path.join(PdParser.__HVLIB_CONVERTED_DIR, obj_type + ".hv.json")):
                            self.obj_counter[obj_type] += 1
                            hv_path = os.path.join(PdParser.__HVLIB_CONVERTED_DIR, obj_type + ".hv.json")
                            x = HeavyGraph(
                                hv_path=hv_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        # is this object in lib/pd?
                        elif os.path.isfile(os.path.join(PdParser.__PDLIB_DIR, obj_type + ".pd")):
                            self.obj_counter[obj_type] += 1
                            pdlib_path = os.path.join(PdParser.__PDLIB_DIR, obj_type + ".pd")

                            # mapping of pd/lib abstraction objects to classes
                            # for checking connection validity
                            clazz = {
                                "abs~": PdLibSignalGraph,
                                "clip~": PdLibSignalGraph,
                                "cos~": PdLibSignalGraph,
                                "dbtopow~": PdLibSignalGraph,
                                "dbtorms~": PdLibSignalGraph,
                                "exp~": PdLibSignalGraph,
                                "ftom~": PdLibSignalGraph,
                                "hip~": PdLibSignalGraph,
                                "lop~": PdLibSignalGraph,
                                "mtof~": PdLibSignalGraph,
                                "powtodb~": PdLibSignalGraph,
                                "q8_rsqrt~": PdLibSignalGraph,
                                "q8_sqrt~": PdLibSignalGraph,
                                "rmstodb~": PdLibSignalGraph,
                                "rsqrt~": PdLibSignalGraph,
                                "sqrt~": PdLibSignalGraph,
                                "wrap~": PdLibSignalGraph
                            }.get(obj_type, PdGraph)

                            x = self.graph_from_file(
                                file_path=pdlib_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False,
                                pd_graph_class=clazz)

                            # register any object-specific warnings or errors
                            if obj_type in ["rzero~", "rzero_rev~", "czero~", "czero_rev~"]:
                                g.add_warning(
                                    "[{0}] accepts only signal input. "
                                    "Arguments and control connections are ignored.".format(obj_type))

                        # is this object in lib/heavy?
                        elif os.path.isfile(os.path.join(PdParser.__HVLIB_DIR, obj_type + ".pd")):
                            self.obj_counter[obj_type] += 1
                            hvlib_path = os.path.join(PdParser.__HVLIB_DIR, obj_type + ".pd")
                            x = self.graph_from_file(
                                file_path=hvlib_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False)

                        # is this an object that must be programatically parsed?
                        elif obj_type in PdParser.__PD_CLASSES:
                            self.obj_counter[obj_type] += 1
                            obj_class = PdParser.__PD_CLASSES[obj_type]
                            x = obj_class(
                                obj_type,
                                obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        elif PdParser.__is_float(obj_type):
                            # parse float literals
                            self.obj_counter["float"] += 1
                            x = HeavyObject(
                                obj_type="__var",
                                obj_args=[float(obj_type)],
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        else:
                            g.add_error(
                                "Don't know how to parse object \"{0}\". Is it an "
                                "object supported by Heavy? Is it an abstraction? "
                                "Have the search paths been correctly configured?".format(obj_type),
                                NotificationEnum.ERROR_UNKNOWN_OBJECT)
                            x = HeavyObject(
                                obj_type="comment",
                                obj_args=["null object placeholder ({0})".format(obj_type)])

                        g.add_object(x)

                    elif line[1] in ["floatatom", "symbolatom"]:
                        self.obj_counter[line[1]] += 1
                        x = self.graph_from_file(
                            file_path=os.path.join(PdParser.__PDLIB_DIR, line[1] + ".pd"),
                            obj_args=[],
                            pos_x=int(line[2]), pos_y=int(line[3]),
                            is_root=False)
                        g.add_object(x)

                    elif line[1] == "array":
                        assert obj_array is None, "#X array object is already being parsed."
                        # array names can have dollar arguments in them.
                        # ensure that they are resolved
                        table_name = PdParser.__resolve_object_args(
                            obj_type="array",
                            obj_args=[line[2]],
                            graph=g)[0]
                        # Pd encodes arrays with length greater than 999,999 with
                        # scientific notatation (e.g. 1e6) which Python's int() can't parse
                        table_size = int(decimal.Decimal(line[3]))
                        obj_array = HeavyObject(
                            obj_type="table",
                            # ensure that obj_array has its own values instance
                            obj_args=[table_name, table_size, []])
                        # TODO(mhroth): position information
                        g.add_object(obj_array)

                    elif line[1] == "msg":
                        self.obj_counter["msg"] += 1
                        g.add_object(PdMessageObject(
                            obj_type="msg",
                            obj_args=[" ".join(line[4:])],
                            pos_x=int(line[2]),
                            pos_y=int(line[3])))

                    elif line[1] == "connect":
                        g.add_parsed_connection(
                            from_index=int(line[2]),
                            from_outlet=int(line[3]),
                            to_index=int(line[4]),
                            to_inlet=int(line[5]))

                    elif line[1] == "declare":
                        if not is_root:
                            g.add_warning(
                                "[declare] objects are not supported in abstractions. "
                                "They can only be in the root canvas.")
                        elif len(line) >= 4 and line[2] == "-path":
                            did_add = self.add_relative_search_directory(line[3])
                            if not did_add:
                                g.add_warning(
                                    "\"{0}\" is not a valid relative abstraction "
                                    "search path. It will be ignored.".format(line[3]))

                        else:
                            g.add_warning(
                                "Heavy only supports the -path flag for the "
                                "declare object.",
                                NotificationEnum.WARNING_DECLARE_PATH)

                    elif line[1] == "coords":
                        pass  # don't do anything with this command

                    else:
                        g.add_error("Don't know how to parse line: {0}".format(
                            " ".join(line)))

                elif line[0] == "#A":
                    obj_array.obj_args["values"].extend([float(f) for f in line[2:]])

                else:
                    g.add_error("Don't know how to parse line: {0}".format(" ".join(line)))

        except Exception as e:
            # bubble the Exception back to the root graph where the graph
            # will be returned
            if not g.is_root:
                raise e
            else:
                # NOTE(mhroth): should the exception be added as an error?
                # Sometimes it's all that we have, so perhaps it's a good idea.
                g.add_error(str(e), NotificationEnum.ERROR_EXCEPTION)

        return g

    @classmethod
    def __resolve_object_args(clazz, obj_type, obj_args, graph, raise_on_failure=True, is_root=False):
        """ Resolve object arguments against the given graph arguments.
            By default this function raises an Exception if it cannot resolve an
            argument.
            This behaviour may be disabled, in which case the unresolved argument
            is replaced with None (which is an otherwise invalid value). A value of
            None typically indicates to a HeavyObject that the default value
            may be used.
        """
        # TODO(mhroth): can this be done more elegantly?
        resolved_obj_args = list(obj_args)  # make a copy of the original obj_args
        for i, a in enumerate(obj_args):
            for m in set(PdParser.__RE_DOLLAR.findall(a)):
                x = int(m)  # the dollar index (i.e. $x)
                if len(graph.obj_args) > x:
                    a = a.replace("\$" + m, str(graph.obj_args[x]))

                # check if hv_args can be used to supply a default value
                elif len(graph.hv_args) > (x - 1):  # heavy args are zero-indexed
                    if not graph.hv_args[x - 1]["required"]:
                        a = a.replace("\$" + m, str(graph.hv_args[x - 1]["default"]))
                    else:
                        graph.add_error(
                            "There is a missing required argument named \"{0}\".".format(
                                graph.hv_args[x - 1]["name"]),
                            NotificationEnum.ERROR_MISSING_REQUIRED_ARGUMENT)

                elif is_root:
                    # NOTE(mhroth): this case is questionable, but since Pd
                    # defaults to this behavior without warning, so will we.
                    # graph.add_warning(
                    #     "${0} in \"{1}\" in the top-level graph is resolved to "
                    #     "\"0\". It is recommended that you remove $-arguments "
                    #     "from the top-level graph.".format(m, a))
                    a = a.replace("\$" + m, "0")

                else:
                    if raise_on_failure:
                        # NOTE(mhroth): this case is questionable, but since Pd
                        # defaults to this behavior without warning, so will we.
                        # graph.add_warning(
                        #     "Object [{0}] requires argument \"{1}\" but the parent "
                        #     "patch does not provide one ({2}). A default value of "
                        #     "\"0\" will be used.".format(obj_type, a, graph.obj_args))
                        a = a.replace("\$" + m, "0")
                    else:
                        a = None  # indicate that this argument could not be resolved by replacing it with None
            resolved_obj_args[i] = a
        return resolved_obj_args

    # a mapping of Pd objects to the classes that will parse them
    __PD_CLASSES = {
        "adc~": PdAudioIoObject,
        "dac~": PdAudioIoObject,
        "inlet": PdLetObject,
        "inlet~": PdLetObject,
        "outlet": PdLetObject,
        "outlet~": PdLetObject,
        "pack": PdPackObject,
        "route": PdRouteObject,
        "sel": PdSelectObject,
        "select": PdSelectObject,
        "t": PdTriggerObject,
        "trigger": PdTriggerObject,
        "table": PdTableObject,
        "unpack": PdUnpackObject,
        "s": PdSendObject,
        "send": PdSendObject,
        "s~": PdSendObject,
        "send~": PdSendObject,
        "throw~": PdSendObject,
        "r": PdReceiveObject,
        "receive": PdReceiveObject,
        "r~": PdReceiveObject,
        "receive~": PdReceiveObject,
        "catch~": PdReceiveObject
    }

    # fill in as much of __PD_CLASSES programmatically as possible
    for o in PdBinopObject.get_supported_objects():
        __PD_CLASSES[o] = PdBinopObject

    @classmethod
    def __is_float(clazz, x):
        """ Returns True if the input can be converted to a float. False otherwise.
        """
        try:
            float(x)
            return True
        except Exception:
            return False



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalPhasor(HeavyObject):

    c_struct = "SignalPhasor"
    preamble = "sPhasor"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvSignalPhasor.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvSignalPhasor.h", "HvSignalPhasor.c"}

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        if obj_type == "__phasor~f":
            return [
                "sPhasor_init(&sPhasor_{0}, sampleRate);".format(obj_id)
            ]
        elif obj_type == "__phasor_k~f":
            return [
                "sPhasor_k_init(&sPhasor_{0}, {1}f, sampleRate);".format(
                    obj_id,
                    args["frequency"])
            ]
        else:
            raise Exception()

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        if obj_type == "__phasor~f":
            return [
                "sPhasor_onMessage(_c, &Context(_c)->sPhasor_{0}, {1}, m);".format(
                    obj_id,
                    inlet_index)]
        elif obj_type == "__phasor_k~f":
            return [
                "sPhasor_k_onMessage(_c, &Context(_c)->sPhasor_{0}, {1}, m);".format(
                    obj_id,
                    inlet_index)]
        else:
            raise Exception()

    @classmethod
    def get_C_process(clazz, obj_type, process_dict, args):
        if obj_type == "__phasor~f":
            return [
                "__hv_phasor_f(&sPhasor_{0}, VIf({1}), VOf({2}));".format(
                    process_dict["id"],
                    HeavyObject._c_buffer(process_dict["inputBuffers"][0]),
                    HeavyObject._c_buffer(process_dict["outputBuffers"][0])
                )
            ]
        elif obj_type == "__phasor_k~f":
            return [
                "__hv_phasor_k_f(&sPhasor_{0}, VOf({1}));".format(
                    process_dict["id"],
                    HeavyObject._c_buffer(process_dict["outputBuffers"][0])
                )
            ]
        else:
            raise Exception("Unknown object type \"{0}\".".format(obj_type))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import os
import time

from .PdParser import PdParser


class Colours:
    purple = "\033[95m"
    cyan = "\033[96m"
    dark_cyan = "\033[36m"
    blue = "\033[94m"
    green = "\033[92m"
    yellow = "\033[93m"
    red = "\033[91m"
    bold = "\033[1m"
    underline = "\033[4m"
    end = "\033[0m"


class pd2hv:

    @classmethod
    def get_supported_objects(clazz):
        return PdParser.get_supported_objects()

    @classmethod
    def compile(clazz, pd_path, hv_dir, search_paths=None, verbose=False, export_args=False):
        tick = time.time()

        parser = PdParser()  # create parser state
        if search_paths is not None:
            for p in search_paths:
                parser.add_absolute_search_directory(p)

        pd_graph = parser.graph_from_file(pd_path)
        notices = pd_graph.get_notices()

        # check for errors
        if len(notices["errors"]) > 0:
            return {
                "stage": "pd2hv",
                "obj_counter": dict(parser.obj_counter),
                "notifs": {
                    "has_error": True,
                    "exception": None,
                    "errors": notices["errors"],
                    "warnings": notices["warnings"]
                },
                "in_dir": os.path.dirname(pd_path),
                "in_file": os.path.basename(pd_path),
                "out_dir": None,
                "out_file": None,
                "compile_time": (time.time() - tick)
            }

        if not os.path.exists(hv_dir):
            os.makedirs(hv_dir)

        hv_file = os.path.splitext(os.path.basename(pd_path))[0] + ".hv.json"
        hv_path = os.path.join(hv_dir, hv_file)
        with open(hv_path, "w") as f:
            if verbose:
                json.dump(pd_graph.to_hv(export_args=export_args),
                          f,
                          sort_keys=True,
                          indent=2,
                          separators=(",", ": "))
            else:
                json.dump(pd_graph.to_hv(export_args=export_args), f)

        return {
            "stage": "pd2hv",
            "obj_counter": dict(parser.obj_counter),
            "notifs": {
                "has_error": False,
                "exception": None,
                "errors": notices["errors"],
                "warnings": notices["warnings"]
            },
            "in_dir": os.path.dirname(pd_path),
            "in_file": os.path.basename(pd_path),
            "out_dir": hv_dir,
            "out_file": hv_file,
            "compile_time": (time.time() - tick)
        }


def main():
    parser = argparse.ArgumentParser(
        description="Converts a Pd patch into the Heavy language format.")
    parser.add_argument(
        "pd_path",
        help="The Pd patch to convert to Heavy.")
    parser.add_argument(
        "hv_dir",
        help="Directory to store generated Heavy patches.")
    parser.add_argument(
        "--export",
        help="Export Heavy arguments. Use this to make precompiled patches.",
        action="count")
    parser.add_argument(
        "-v",
        "--verbose",
        help="Show debugging information.",
        action="count")
    args = parser.parse_args()

    args.pd_path = os.path.abspath(os.path.expanduser(args.pd_path))
    args.hv_dir = os.path.abspath(os.path.expanduser(args.hv_dir))

    result = pd2hv.compile(
        pd_path=args.pd_path,
        hv_dir=args.hv_dir,
        search_paths=None,
        verbose=args.verbose,
        export_args=args.export)

    for i, n in enumerate(result["notifs"]["errors"]):
        print("{0:3d}) {1}Error #{2:4d}:{3} {4}".format(
            i + 1,
            Colours.red,
            n["enum"],
            Colours.end,
            n["message"]))
    for i, n in enumerate(result["notifs"]["warnings"]):
        print("{0:3d}) {1}Warning #{2:4d}:{3} {4}".format(
            i + 1,
            Colours.yellow,
            n["enum"],
            Colours.end,
            n["message"]))

    if args.verbose:
        if len(result["notifs"]["errors"]) == 0:
            print("Heavy file written to", os.path.join(result["out_dir"], result["out_file"]))
        print("Total pd2hv compile time: {0:.2f}ms".format(result["compile_time"] * 1000))


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalTabwrite(HeavyObject):
    """Handles __tabwrite~f
    """

    c_struct = "SignalTabwrite"
    preamble = "sTabwrite"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvSignalTabwrite.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvSignalTabwrite.h", "HvSignalTabwrite.c"}

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return [
            "sTabwrite_init(&sTabwrite_{0}, &hTable_{1});".format(
                obj_id,
                args["table_id"])]

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return [
            "sTabwrite_onMessage(_c, &Context(_c)->sTabwrite_{0}, {1}, m, NULL);".format(
                obj_id,
                inlet_index)
        ]

    @classmethod
    def get_C_process(clazz, obj_type, process_dict, objects):
        if obj_type == "__tabwrite~f":
            return [
                "__hv_tabwrite_f(&sTabwrite_{0}, {1});".format(
                    process_dict["id"],
                    ", ".join(["VIf({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["inputBuffers"]])
                )]
        elif obj_type == "__tabwrite_stoppable~f":
            return [
                "__hv_tabwrite_stoppable_f(&sTabwrite_{0}, {1});".format(
                    process_dict["id"],
                    ", ".join(["VIf({0})".format(HeavyObject._c_buffer(b)) for b in process_dict["inputBuffers"]])
                )]
        else:
            raise Exception()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os

from .HIrConvolution import HIrConvolution
from .HIrInlet import HIrInlet
from .HIrLorenz import HIrLorenz
from .HIrOutlet import HIrOutlet
from .HIrPack import HIrPack
from .HIrSwitchcase import HIrSwitchcase
from .HIrTabhead import HIrTabhead
from .HIrTabread import HIrTabread
from .HIrTabwrite import HIrTabwrite

from .HLangAdc import HLangAdc
from .HLangBinop import HLangBinop
from .HLangBiquad import HLangBiquad
from .HLangDelay import HLangDelay
# from .HLangIf import HLangIf
from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject
from .HLangLine import HLangLine
from .HLangMessage import HLangMessage
from .HLangPhasor import HLangPhasor
from .HLangPrint import HLangPrint
from .HLangReceive import HLangReceive
from .HLangRandom import HLangRandom
from .HLangSend import HLangSend
from .HLangSequence import HLangSequence
from .HLangSlice import HLangSlice
from .HLangSystem import HLangSystem
from .HLangTable import HLangTable
from .HLangUnop import HLangUnop
from .HLangVar import HLangVar
from .HLangVario import HLangVario

from .HeavyGraph import HeavyGraph
from .Connection import Connection


class HeavyParser:

    @classmethod
    def graph_from_file(clazz, hv_file, graph=None, graph_args=None, path_stack=None, xname=None):
        """ Read a graph object from a file.

            @param graph  The parent graph of this graph.
            @param graph_args  The arguments to this graph, in the form of a completely
            resolved dictionary.
            @param path_stack  The path_stack is the current stack of resolved abstractions.
            It prevents infinite recursion when reading many abstractions deep.
        """
        # ensure that we have an absolute path to the hv_file
        hv_file = os.path.abspath(os.path.expanduser(hv_file))

        # copy the path stack such that no changes are made to the calling stack
        path_stack = path_stack or set()
        if hv_file in path_stack:
            raise HeavyException("Abstraction recursion detected. Rereading {0} on stack {1}.".format(
                hv_file,
                path_stack))
        else:
            path_stack.add(hv_file)

        # open and parse the heavy file
        with open(hv_file, "r") as f:
            json_heavy = json.load(f)

        return HeavyParser.graph_from_object(json_heavy, graph, graph_args, hv_file, path_stack, xname)

    @classmethod
    def graph_from_object(clazz, json_heavy, graph=None, graph_args=None, hv_file=None, path_stack=None, xname=None):
        """ Parse a graph object.

            @param graph  The parent graph.
            @param graph_args  The resolved arguments to this graph, in the form of a dictionary.
            @param hv_file  The Heavy file where this graph can be found.
        """
        # resolve default graph arguments
        graph_args = graph_args or {}
        for a in json_heavy["args"]:
            if a["name"] not in graph_args:
                if a["required"]:
                    raise HeavyException("Required argument \"{0}\" not present.".format(a["name"]))
                else:
                    graph_args[a["name"]] = a["default"]
            else:
                # just to be safe, ensure that the argument type is correct
                graph_args[a["name"]] = HeavyLangObject.force_arg_type(
                    graph_args[a["name"]],
                    a["value_type"],
                    graph=graph)

        # create a new graph
        subpatch_name = json_heavy.get("annotations", {}).get("name", xname)
        g = HeavyGraph(graph, graph_args, file=hv_file, xname=subpatch_name)

        # add the import paths to the global vars
        g.local_vars.add_import_paths(json_heavy.get("imports", []))
        # add the file's relative directory to global vars
        g.local_vars.add_import_paths([os.path.dirname(hv_file)])

        # instantiate all objects
        try:
            for obj_id, o in json_heavy["objects"].items():
                if o["type"] == "comment":
                    continue  # first and foremost, ignore comment objects

                elif o["type"] == "graph":
                    # inline HeavyGraph objects (i.e. subgraphs)
                    # require a different set of initialisation arguments
                    x = HeavyParser.graph_from_object(o, g, g.args, hv_file, path_stack, xname)

                else:
                    # resolve the arguments dictionary based on the graph args
                    args = g.resolve_arguments(o["args"])

                    # before anything, search for an abstraction
                    # in case we want to override default functionality
                    # However, if we are in an abstraction that has the same
                    # name as the type that we are looking for, don't recurse!
                    abs_path = g.find_path_for_abstraction(o["type"])
                    if abs_path is not None and abs_path not in path_stack:
                        x = HeavyParser.graph_from_file(
                            hv_file=abs_path,
                            graph=g,
                            graph_args=args,
                            path_stack=path_stack)

                    # if we know how to handle this object type natively
                    # either as a custom type or as a generic IR object
                    elif HeavyParser.get_class_for_type(o["type"]) is not None:
                        obj_clazz = HeavyParser.get_class_for_type(o["type"])
                        x = obj_clazz(o["type"], args, g, o.get("annotations", {}))

                    # handle generic IR objects
                    elif HeavyIrObject.is_ir(o["type"]):
                        x = HeavyIrObject(o["type"], args, g, annotations=o.get("annotations", {}))

                    # an object definition can't be found
                    else:
                        g.add_error("Object type \"{0}\" cannot be found.".format(o["type"]))
                        # note that add_error() raises an exception. So really, there is no continue.
                        continue

                # add the new object to the graph's object dictionary
                g.add_object(x, obj_id)

            # parse all of the connections
            for c in json_heavy["connections"]:
                g.connect_objects(Connection(
                    g.objs[c["from"]["id"]],
                    c["from"]["outlet"],
                    g.objs[c["to"]["id"]],
                    c["to"]["inlet"],
                    c["type"]))

        except HeavyException as e:
            if g.is_root_graph():
                # add the notification dictionary at the top level
                e.notes = g.get_notices()
                e.notes["has_error"] = True
                e.notes["exception"] = e
            raise e

        if (g.graph is None) or (g.graph.file != g.file):
            # remove this graph from the stack when finished.
            # Subpatches should not remove themselves.
            path_stack.remove(g.file)

        return g

    @classmethod
    def get_class_for_type(clazz, obj_type):
        """ Returns the class which can handle the given object type.
        """
        if HLangUnop.handles_type(obj_type):
            return HLangUnop
        elif HLangBinop.handles_type(obj_type):
            return HLangBinop
        elif obj_type in LANG_CLASS_DICT:
            return LANG_CLASS_DICT[obj_type]
        else:
            return None


class HLangIf(HeavyLangObject):
    """ Translates HeavyLang object [if] to HeavyIR [if] or [if~].
    """

    def __init__(self, obj_type, args, graph, annotations=None):
        HeavyLangObject.__init__(self, "if", args, graph,
                                 num_inlets=2,
                                 num_outlets=2,
                                 annotations=annotations)

    def reduce(self):
        if self.has_inlet_connection_format(["cc", "_c", "c_", "__"]):
            x = HeavyIrObject("__if", self.args)
        elif self.has_inlet_connection_format("ff"):
            # TODO(mhroth): implement this
            x = HeavyParser.graph_from_file("./hvlib/if~f.hv.json")
        elif self.has_inlet_connection_format("ii"):
            # TODO(mhroth): implement this
            x = HeavyParser.graph_from_file("./hvlib/if~i.hv.json")
        else:
            raise HeavyException("Unhandled connection configuration to object [if]: {0}".format(
                self._get_connection_format(self.inlet_connections)))

        return ({x}, self.get_connection_move_list(x))


# NOTE(mhroth): these imports are at the end of the file in order to prevent
# circular import errors
from .HLangDac import HLangDac
from .HLangNoise import HLangNoise

# A list of all of the HeavyLang objects and the classes
# that will translate them into HeavyIR objects.
LANG_CLASS_DICT = {
    "__conv~f": HIrConvolution,
    "biquad": HLangBiquad,
    "if": HLangIf,
    "inlet": HIrInlet,
    "var": HLangVar,
    "vario": HLangVario,
    "outlet": HIrOutlet,
    "__lorenz~f": HIrLorenz,
    "print": HLangPrint,
    "sequence": HLangSequence,
    "adc": HLangAdc,
    "dac": HLangDac,
    "message": HLangMessage,
    "noise": HLangNoise,
    "system": HLangSystem,
    "phasor": HLangPhasor,
    "line": HLangLine,
    "random": HLangRandom,
    "delay": HLangDelay,
    "table": HLangTable,
    "slice": HLangSlice,
    "__tabread~if": HIrTabread,
    "__tabread~f": HIrTabread,
    "__tabreadu~f": HIrTabread,
    "__tabread": HIrTabread,
    "__tabhead~f": HIrTabhead,
    "__tabhead": HIrTabhead,
    "__tabwrite~f": HIrTabwrite,
    "__tabwrite_stoppable~f": HIrTabwrite,
    "__tabwrite": HIrTabwrite,
    "receive": HLangReceive,
    "send": HLangSend,
    "__switchcase": HIrSwitchcase,
    "switchcase": HIrSwitchcase,
    "__pack": HIrPack
}



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject


class HIrInlet(HeavyIrObject):
    """ A specific implementation of the inlet object.
    """

    def __init__(self, obj_type, args=None, graph=None, annotations=None):
        HeavyIrObject.__init__(self, "__inlet", args=args, graph=graph, annotations=annotations)

    def _resolved_outlet_type(self, outlet_index=0):
        connections = self.graph.inlet_connections[self.args["index"]]
        connection_type_set = {c.type for c in connections}
        if len(connection_type_set) == 0:
            # object has no incident connections.
            return "-->"  # outlet type defaults to control (-->)
        elif len(connection_type_set) == 1:
            return list(connection_type_set)[0]
        else:
            raise HeavyException(
                "{0} has multiple incident connections of differing type. "
                "The outlet type cannot be explicitly resolved.".format(self))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class ControlVar(HeavyObject):
    """An object which holds a variable. In this case only a float.
    NOTE(mhroth): maybe in the future this can hold any data structure, such as
    a generic message. At the moment, the memory churn is deemed unnecessary.
    """

    c_struct = "ControlVar"
    preamble = "cVar"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvControlVar.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvControlVar.h", "HvControlVar.c"}

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        if isinstance(args["k"], str):
            return [
                "cVar_init_s(&cVar_{0}, \"{1}\");".format(
                    obj_id,
                    args["k"])]
        else:
            return [
                "cVar_init_f(&cVar_{0}, {1}f);".format(
                    obj_id,
                    float(args["k"]))]

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return [
            "cVar_onMessage(_c, &Context(_c)->cVar_{0}, {1}, m, &cVar_{0}_sendMessage);".format(
                obj_id,
                inlet_index)
        ]



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

class Connection:
    def __init__(self, from_obj_id, outlet_index, to_obj_id, inlet_index, conn_type):
        self.__hv_json = {
            "from": {
                "id": from_obj_id,
                "outlet": outlet_index
            },
            "to": {
                "id": to_obj_id,
                "inlet": inlet_index
            },
            "type": conn_type
        }

    def to_hv(self):
        return self.__hv_json



# Copyright 2015 Enzien Audio, Ltd. All Rights Reserved.

import json
import os

from .HeavyObject import HeavyObject
from .MaxAdcObject import MaxAdcObject
from .MaxBinopObject import MaxBinopObject
from .MaxDacObject import MaxDacObject
from .MaxGraph import MaxGraph
from .MaxInletObject import MaxInletObject
from .MaxOutletObject import MaxOutletObject
from .MaxUnopObject import MaxUnopObject


class MaxParser:

    @classmethod
    def graph_from_file(clazz, file_path, obj_id=None, obj_args=None):
        with open(file_path, "r") as f:
            graph_obj = json.loads(f.read())

        return MaxParser.graph_from_object(graph_obj, obj_id, obj_args, file_path)

    @classmethod
    def graph_from_object(clazz, obj, obj_id=None, obj_args=None, max_path=None):
        g = MaxGraph(obj_args=obj_args, obj_id=obj_id, max_path=max_path)

        # parse objects
        for o in obj["patcher"]["boxes"]:
            o = o["box"]  # the max object dictionary
            x = None  # a MaxObject

            # read common parameters
            obj_id = o["id"]
            patching_rect = o["patching_rect"]

            if o["maxclass"] == "newobj":
                obj_text = o["text"].split()
                obj_type = obj_text[0]
                obj_args = obj_text[1:] if len(obj_text) > 1 else []

                # resolve object arguments against graph arguments
                for i, a in enumerate(obj_args):
                    if a.startswith("$"):
                        x = int(a[1:]) - 1
                        if len(g.obj_args) > x:
                            obj_args[i] = g.obj_args[x]
                        else:
                            print("WARNING: Can't resolve object argument {0} against graph arguments: {1}".format(
                                a,
                                g.obj_args))

                # are we dealing with a declared Heavy object?
                if obj_type == "patcher":
                    if len(obj_args) > 0 and obj_args[0] == "@hv_obj":
                        if len(obj_args) > 1 and HeavyObject.is_heavy(obj_args[1]):
                            # initialise heavy object
                            x = HeavyObject(
                                obj_args[1], obj_args[2:], obj_id,
                                pos_x=patching_rect[0], pos_y=patching_rect[1])
                        else:
                            raise Exception("Max subpatch has @hv_obj annotation but declares no Heavy object.")
                    else:
                        # TODO(mhroth): assume that this is a normal max subpatch
                        raise Exception("Cannot parse Max subpatches yet.")

                # do we have an abstraction for this max object? Is it in the maxlib?
                # TODO(mhroth): are there any other search paths to look through?
                elif os.path.isfile(os.path.join(os.path.dirname(__file__), "maxlib", obj_type + ".maxpat")):
                    max_path = os.path.join(os.path.dirname(__file__), "maxlib", obj_type + ".maxpat")
                    x = MaxParser.graph_from_file(max_path, obj_id, obj_args)

                # is this an object that must be programmatically parsed?
                elif obj_type in MaxParser.__MAX_CLASSES:
                    clazz = MaxParser.__MAX_CLASSES[obj_type]
                    x = clazz(
                        obj_type, obj_args, obj_id,
                        pos_x=patching_rect[0], pos_y=patching_rect[1])

                else:
                    raise Exception("Don't know how to parse object \"{0}\".".format(obj_type))

            elif o["maxclass"] == "inlet":
                outlet_type = o["outlettype"][0]
                x = MaxInletObject(
                    outlet_type, obj_id,
                    pos_x=patching_rect[0], pos_y=patching_rect[1])

            elif o["maxclass"] == "outlet":
                # TODO(mhroth): correctly determine connection type of outlet
                x = MaxOutletObject(
                    "signal", obj_id,
                    pos_x=patching_rect[0], pos_y=patching_rect[1])

            else:
                print("WARNING: Ignoring maxclass \"{0}\".".format(o["maxclass"]))
                continue  # ignore comments, etc.

            g.add_object(x)

        # parse connections
        for li in obj["patcher"]["lines"]:
            li = li["patchline"]
            if not (bool(li["disabled"]) or bool(li["hidden"])):
                g.add_connection(
                    from_id=li["source"][0],
                    from_outlet=li["source"][1],
                    to_id=li["destination"][0],
                    to_inlet=li["destination"][1])

        return g

    # a mapping of Max objects to the classes that will parse them
    __MAX_CLASSES = {
        "adc~": MaxAdcObject,
        "dac~": MaxDacObject
    }
    # fill in as much of __MAX_CLASSES programmatically as possible
    for o in MaxUnopObject.get_supported_objects():
        __MAX_CLASSES[o] = MaxUnopObject
    for o in MaxBinopObject.get_supported_objects():
        __MAX_CLASSES[o] = MaxBinopObject



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
import random
import string

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum


class PdObject:

    __RANDOM = random.Random()
    __ID_CHARS = string.ascii_letters + string.digits

    def __init__(self, obj_type, obj_args=None, pos_x=0, pos_y=0):
        self.obj_type = obj_type
        # all arguments should be resolved when passed to a PdObject
        self.obj_args = obj_args or []
        self.obj_id = "{0}_{1}".format(
            obj_type,
            "".join(PdObject.__RANDOM.choice(PdObject.__ID_CHARS) for _ in range(8)))
        self.pos_x = pos_x
        self.pos_y = pos_y

        # this is set when the object is added to a graph
        self.parent_graph = None

        self._inlet_connections = defaultdict(list)
        self._outlet_connections = defaultdict(list)

        self._warnings = []
        self._errors = []

    def add_warning(self, warning, enum=NotificationEnum.WARNING_GENERIC):
        """ Add a warning regarding this object.
        """
        self._warnings.append({"enum": enum, "message": warning})

    def add_error(self, error, enum=NotificationEnum.ERROR_GENERIC):
        """ Add an error regarding this object.
        """
        self._errors.append({"enum": enum, "message": error})

    def get_notices(self):
        """ Returns a dictionary of all warnings and errors at this object.
        """
        # TODO(mhroth): we might want to consider moving to a more expressive format.
        # {
        #     "objectType": "trigger",
        #     "objectString": "t l l",
        #     "graphs": ["_main", "hello"],
        #     "rawText": "Heavy only supports arguments 'a', 'f', 's', and 'b'.",
        #     "humanText": "[t l l] in "_main/osc~" @ (x:452, y:273): Heavy only supports
        # arguments 'a', 'f', 's', and 'b'.",
        #     "position": {
        #         "x": 452,
        #         "y": 273
        #     }
        # }
        """
        {
            "warnings": [{
                "enum": 1000,
                "message": "this is a warning"
            }],
            "errors": [{
                "enum": 2000,
                "message": "this is an error"
            }]
        }
        """
        return {
            "warnings": [
                {
                    "enum": n["enum"],
                    "message": "{0} in \"{1}\" @ (x:{2}, y:{3}): {4}".format(
                        self, "/".join(self.get_graph_heirarchy()), self.pos_x, self.pos_y, n["message"])
                } for n in self._warnings
            ],
            "errors": [
                {
                    "enum": n["enum"],
                    "message": "{0} in \"{1}\" @ (x:{2}, y:{3}): {4}".format(
                        self, "/".join(self.get_graph_heirarchy()), self.pos_x, self.pos_y, n["message"])
                } for n in self._errors
            ]
        }

    def get_inlet_connection_type(self, inlet_index=0):
        """ Returns the inlet connection type of this Pd object.
            For the sake of convenience, the connection type is reported in
            Heavy's format.
        """
        return "~f>" if self.obj_type.endswith("~") else "-->"

    def get_outlet_connection_type(self, outlet_index=0):
        """ Returns the outlet connection type of this Pd object.
            For the sake of convenience, the connection type is reported in
            Heavy's format.
        """
        return "~f>" if self.obj_type.endswith("~") else "-->"

    def add_connection(self, c):
        """ Adds a connection, either inlet or outlet, to this object.
        """
        if c.from_id == self.obj_id:
            self._outlet_connections[str(c.outlet_index)].append(c)
        elif c.to_id == self.obj_id:
            self._inlet_connections[str(c.inlet_index)].append(c)
        else:
            raise Exception("Adding a connection to the wrong object!")

    def remove_connection(self, c):
        """ Remove a connection to this object.
        """
        if c.to_obj is self:
            self._inlet_connections[str(c.inlet_index)].remove(c)
        elif c.from_obj is self:
            self._outlet_connections[str(c.outlet_index)].remove(c)
        else:
            raise Exception("Connection {0} does not connect to this object {1}.".format(c, self))

    def get_graph_heirarchy(self):
        """ Returns an indication of the graph "path" of this object.
        It only includes unique graphs (not subpatches) E.g. _main/tabosc4~
        The check for None is in case the object is somehow not yet attached.
        """
        return self.parent_graph.get_graph_heirarchy() \
            if self.parent_graph is not None else ["unattached"]

    def validate_configuration(self):
        """ Called when all graphs are finished parsing, from the root.
            Gives each object the chance to validate it's configuration,
            including connections.
            In general this function does nothing, though it may add warnings and
            errors.
            Note that object validation occurs when the entire patch is finished
            parsing and before it is returned to the user. The program may assume
            that all parameters and variables have been set.
        """
        pass

    @classmethod
    def get_supported_objects(clazz):
        """ Returns a list of Pd objects that this class can parse.
        """
        raise NotImplementedError()

    def to_hv(self):
        """ Returns the HeavyLang JSON representation of this object.
        """
        raise NotImplementedError()

    def __repr__(self):
        if len(self.obj_args) == 0:
            return "[{0}]".format(self.obj_type)
        else:
            return "[{0} {1}]".format(
                self.obj_type,
                " ".join([str(o) for o in self.obj_args]))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalEnvelope(HeavyObject):

    c_struct = "SignalEnvelope"
    preamble = "sEnv"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvSignalEnvelope.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvSignalEnvelope.h", "HvSignalEnvelope.c"}

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return [
            "sEnv_init(&sEnv_{0}, {1}, {2});".format(
                obj_id,
                int(args["windowSize"]),
                int(args["period"]))
        ]

    @classmethod
    def get_C_process(clazz, obj_type, process_dict, args):
        return [
            "sEnv_process(this, &sEnv_{0}, VIf({1}), &sEnv_{0}_sendMessage);".format(
                process_dict["id"],
                HeavyObject._c_buffer(process_dict["inputBuffers"][0])
            )
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyIrObject import HeavyIrObject


class HIrSwitchcase(HeavyIrObject):
    """ A specific implementation of the __switchcase object.
    """

    def __init__(self, obj_type, args=None, graph=None, annotations=None):
        HeavyIrObject.__init__(self, "__switchcase",
                               args=args,
                               graph=graph,
                               num_inlets=1,
                               num_outlets=len(args["cases"]) + 1,
                               annotations=annotations)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class SignalMath(HeavyObject):
    """Handles the math objects.
    """

    # translation from operation to function preamble
    __OPERATION_DICT = {
        "__add~f": "__hv_add_f",
        "__add~i": "__hv_add_i",
        "__sub~f": "__hv_sub_f",
        "__sub~i": "__hv_sub_i",
        "__mul~f": "__hv_mul_f",
        "__mul~i": "__hv_mul_i",
        "__div~f": "__hv_div_f",
        "__div~i": "__hv_div_i",
        "__log2~f": "__hv_log2_f",
        "__cos~f": "__hv_cos_f",
        "__acos~f": "__hv_acos_f",
        "__cosh~f": "__hv_cosh_f",
        "__acosh~f": "__hv_acosh_f",
        "__sin~f": "__hv_sin_f",
        "__asin~f": "__hv_asin_f",
        "__sinh~f": "__hv_sinh_f",
        "__asinh~f": "__hv_asinh_f",
        "__tan~f": "__hv_tan_f",
        "__atan~f": "__hv_atan_f",
        "__atan2~f": "__hv_atan2_f",
        "__tanh~f": "__hv_tanh_f",
        "__atanh~f": "__hv_atanh_f",
        "__exp~f": "__hv_exp_f",
        "__pow~f": "__hv_pow_f",
        "__pow~i": "__hv_pow_i",
        "__sqrt~f": "__hv_sqrt_f",
        "__rsqrt~f": "__hv_rsqrt_f",
        "__abs~f": "__hv_abs_f",
        "__abs~i": "__hv_abs_i",
        "__max~f": "__hv_max_f",
        "__max~i": "__hv_max_i",
        "__min~f": "__hv_min_f",
        "__min~i": "__hv_min_i",
        "__neg~f": "__hv_neg_f",
        "__gt~f": "__hv_gt_f",
        "__gt~i": "__hv_gt_i",
        "__gte~f": "__hv_gte_f",
        "__gte~i": "__hv_gte_i",
        "__lt~f": "__hv_lt_f",
        "__lt~i": "__hv_lt_i",
        "__lte~f": "__hv_lte_f",
        "__lte~i": "__hv_lte_i",
        "__neq~f": "__hv_neq_f",
        "__fma~f": "__hv_fma_f",
        "__fms~f": "__hv_fms_f",
        "__floor~f": "__hv_floor_f",
        "__ceil~f": "__hv_ceil_f",
        "__cast~fi": "__hv_cast_fi",
        "__cast~if": "__hv_cast_if",
        "__and~f": "__hv_and_f",  # binary and
        "__andnot~f": "__hv_andnot_f",
        "__or~f": "__hv_or_f",  # binary or
    }

    @classmethod
    def handles_type(clazz, obj_type):
        """Returns true if the object type can be handled by this class
        """
        return obj_type in SignalMath.__OPERATION_DICT

    @classmethod
    def get_C_header_set(clazz):
        return {"HvMath.h"}

    @classmethod
    def get_C_process(clazz, obj_type, process_dict, objects):
        return [
            "{0}({1}, {2});".format(
                SignalMath.__OPERATION_DICT[obj_type],
                ", ".join(["VI{0}({1})".format(
                    "i" if b["type"] == "~i>" else "f",
                    HeavyObject._c_buffer(b)) for b in process_dict["inputBuffers"]]),
                ", ".join(["VO{0}({1})".format(
                    "i" if b["type"] == "~i>" else "f",
                    HeavyObject._c_buffer(b)) for b in process_dict["outputBuffers"]])
            )]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
from collections import OrderedDict
import json
import os
import re
import time

from hvcc.interpreters.pd2hv import pd2hv
from hvcc.interpreters.max2hv import max2hv
from hvcc.core.hv2ir import hv2ir
from hvcc.generators.ir2c import ir2c
from hvcc.generators.ir2c import ir2c_perf
from hvcc.generators.c2bela import c2bela
from hvcc.generators.c2fabric import c2fabric
from hvcc.generators.c2js import c2js
from hvcc.generators.c2dpf import c2dpf
from hvcc.generators.c2pdext import c2pdext
from hvcc.generators.c2wwise import c2wwise
from hvcc.generators.c2unity import c2unity


class Colours:
    purple = "\033[95m"
    cyan = "\033[96m"
    dark_cyan = "\033[36m"
    blue = "\033[94m"
    green = "\033[92m"
    yellow = "\033[93m"
    red = "\033[91m"
    bold = "\033[1m"
    underline = "\033[4m"
    end = "\033[0m"


def add_error(results, error):
    if "hvcc" in results:
        results["hvcc"]["notifs"]["errors"].append({"message": error})
    else:
        results["hvcc"] = {
            "stage": "hvcc",
            "notifs": {
                "has_error": True,
                "exception": None,
                "errors": [{"message": error}],
                "warnings": []
            }
        }
    return results


def check_extern_name_conflicts(extern_type, extern_list, results):
    """ In most of the generator code extern names become capitalised when used
        as enums. This method makes sure that there are no cases where two unique
        keys become the same after being capitalised.
        Note(joe): hvcc is probably the best place to check as at this point we
        have a list of all extern names.
    """
    for i, v in enumerate(extern_list):
        for j, u in enumerate(extern_list[i + 1:]):
            if v[0].upper() == u[0].upper():
                add_error(results,
                          "Conflicting {0} names '{1}' and '{2}', make sure that "
                          "capital letters are not the only difference.".format(extern_type, v[0], u[0]))


def generate_extern_info(hvir, results):
    """ Simplifies the receiver/send and table lists by only containing values
        externed with @hv_param, @hv_event or @hv_table
    """
    # Exposed input parameters
    in_parameter_list = [(k, v) for k, v in hvir["control"]["receivers"].items() if v.get("extern", None) == "param"]
    in_parameter_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("input parameter", in_parameter_list, results)

    # Exposed input events
    in_event_list = [(k, v) for k, v in hvir["control"]["receivers"].items() if v.get("extern", None) == "event"]
    in_event_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("input event", in_event_list, results)

    # Exposed output parameters
    out_parameter_list = [(v["name"], v) for v in hvir["control"]["sendMessage"] if v.get("extern", None) == "param"]
    # remove duplicate output parameters/events
    # NOTE(joe): is the id argument important here? We'll only take the first one in this case.
    out_parameter_list = list(dict(out_parameter_list).items())
    out_parameter_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("output parameter", out_parameter_list, results)

    # Exposed output events
    out_event_list = [(v["name"], v) for v in hvir["control"]["sendMessage"] if v.get("extern", None) == "event"]
    out_event_list = list(dict(out_event_list).items())
    out_event_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("output event", out_event_list, results)

    # Exposed tables
    table_list = [(k, v) for k, v in hvir["tables"].items() if v.get("extern", None)]
    table_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("table", table_list, results)

    return {
        "parameters": {
            "in": in_parameter_list,
            "out": out_parameter_list
        },
        "events": {
            "in": in_event_list,
            "out": out_event_list
        },
        "tables": table_list,
        # generate patch heuristics to ensure enough memory allocated for the patch
        "memoryPoolSizesKb": {
            "internal": 10,  # TODO(joe): should this increase if there are a lot of internal connections?
            "inputQueue": max(2, int(len(in_parameter_list) + len(in_event_list) / 4)),
            "outputQueue": max(2, int(len(out_parameter_list) + len(out_event_list) / 4)),
        }
    }


def compile_dataflow(in_path, out_dir, patch_name=None,
                     search_paths=None, generators=None, verbose=False,
                     copyright=None, hvir=None):

    results = OrderedDict()  # default value, empty dictionary

    # basic error checking on input
    if os.path.isfile(in_path):
        if not in_path.endswith((".pd", ".maxpat")):
            return add_error(results, "Can only process Pd or Max files.")
    elif os.path.isdir(in_path):
        if not os.path.basename("c"):
            return add_error(results, "Can only process c directories.")
    else:
        return add_error(results, "Unknown input path {0}".format(in_path))

    patch_name = patch_name or "heavy"
    generators = generators or {"c"}

    if in_path.endswith((".pd", ".maxpat")):
        if verbose:
            print("--> Generating C")
        if in_path.endswith(".pd"):
            results["pd2hv"] = pd2hv.pd2hv.compile(
                pd_path=in_path,
                hv_dir=os.path.join(out_dir, "hv"),
                search_paths=search_paths,
                verbose=verbose)
        elif in_path.endswith(".maxpat"):
            results["max2hv"] = max2hv.max2hv.compile(
                max_path=in_path,
                hv_dir=os.path.join(out_dir, "hv"),
                search_paths=search_paths,
                verbose=verbose)

        # check for errors
        if list(results.values())[0]["notifs"].get("has_error", False):
            return results

        results["hv2ir"] = hv2ir.hv2ir.compile(
            hv_file=os.path.join(list(results.values())[0]["out_dir"], list(results.values())[0]["out_file"]),
            # ensure that the ir filename has no funky characters in it
            ir_file=os.path.join(out_dir, "ir", re.sub("\W", "_", patch_name) + ".heavy.ir.json"),
            patch_name=patch_name,
            verbose=verbose)

        # check for errors
        if results["hv2ir"]["notifs"].get("has_error", False):
            return results

        # get the hvir data
        hvir = results["hv2ir"]["ir"]
        patch_name = hvir["name"]["escaped"]
        externs = generate_extern_info(hvir, results)

        c_src_dir = os.path.join(out_dir, "c")
        results["ir2c"] = ir2c.ir2c.compile(
            hv_ir_path=os.path.join(results["hv2ir"]["out_dir"], results["hv2ir"]["out_file"]),
            static_dir=os.path.join(os.path.dirname(__file__), "generators/ir2c/static"),
            output_dir=c_src_dir,
            externs=externs,
            copyright=copyright)

        # check for errors
        if results["ir2c"]["notifs"].get("has_error", False):
            return results

        # ir2c_perf
        results["ir2c_perf"] = {
            "stage": "ir2c_perf",
            "obj_counter": ir2c_perf.ir2c_perf.perf(results["hv2ir"]["ir"], verbose=verbose),
            "in_dir": results["hv2ir"]["out_dir"],
            "in_file": results["hv2ir"]["out_file"],
            "notifs": {}
        }

        # reconfigure such that next stage is triggered
        in_path = c_src_dir

    if os.path.isdir(in_path) and os.path.basename(in_path) == "c":
        # the c code is provided
        c_src_dir = in_path
        if hvir is None:
            # if hvir ir not provided, load it from the ir path
            try:
                # hvir_dir == project/c/../ir == project/ir
                hvir_dir = os.path.join(in_path, "..", "ir")
                hvir_path = os.path.join(hvir_dir, os.listdir(hvir_dir)[0])
                if os.path.isfile(hvir_path):
                    with open(hvir_path, "r") as f:
                        hvir = json.load(f)
                        patch_name = hvir["name"]["escaped"]
                        externs = generate_extern_info(hvir, results)
                else:
                    return add_error(results, "Cannot find hvir file.")
            except Exception as e:
                return add_error(results, "ir could not be found or loaded: {0}.".format(e))

    # run the c2x generators, merge the results
    num_input_channels = hvir["signal"]["numInputBuffers"]
    num_output_channels = hvir["signal"]["numOutputBuffers"]

    if "bela" in generators:
        if verbose:
            print("--> Generating Bela plugin")
        results["c2fabric"] = c2bela.c2bela.compile(
            c_src_dir=c_src_dir,
            out_dir=os.path.join(out_dir, "bela"),
            patch_name=patch_name,
            num_input_channels=num_input_channels,
            num_output_channels=num_output_channels,
            externs=externs,
            verbose=verbose)

    if "fabric" in generators:
        if verbose:
            print("--> Generating Fabric plugin")
        results["c2fabric"] = c2fabric.c2fabric.compile(
            c_src_dir=c_src_dir,
            out_dir=os.path.join(out_dir, "fabric"),
            patch_name=patch_name,
            num_input_channels=num_input_channels,
            num_output_channels=num_output_channels,
            externs=externs,
            copyright=copyright,
            verbose=verbose)

    if "js" in generators:
        if verbose:
            print("--> Generating Javascript")
        results["c2js"] = c2js.c2js.compile(
            c_src_dir=c_src_dir,
            out_dir=os.path.join(out_dir, "js"),
            patch_name=patch_name,
            num_input_channels=num_input_channels,
            num_output_channels=num_output_channels,
            externs=externs,
            copyright=copyright,
            verbose=verbose)

    if "dpf" in generators:
        if verbose:
            print("--> Generating DPF plugin")
        results["c2dpf"] = c2dpf.c2dpf.compile(
            c_src_dir=c_src_dir,
            out_dir=os.path.join(out_dir, "distrho"),
            patch_name=patch_name,
            num_input_channels=num_input_channels,
            num_output_channels=num_output_channels,
            externs=externs,
            copyright=copyright,
            verbose=verbose)

    if "pdext" in generators:
        if verbose:
            print("--> Generating Pd external")
        results["c2pdext"] = c2pdext.c2pdext.compile(
            c_src_dir=c_src_dir,
            out_dir=os.path.join(out_dir, "pdext"),
            patch_name=patch_name,
            ext_name=patch_name + "~",
            num_input_channels=num_input_channels,
            num_output_channels=num_output_channels,
            externs=externs,
            copyright=copyright,
            verbose=verbose)

    if "unity" in generators:
        if verbose:
            print("--> Generating Unity plugin")
        results["c2unity"] = c2unity.c2unity.compile(
            c_src_dir=c_src_dir,
            out_dir=os.path.join(out_dir, "unity"),
            patch_name=patch_name,
            num_input_channels=num_input_channels,
            num_output_channels=num_output_channels,
            externs=externs,
            copyright=copyright,
            verbose=verbose)

    if "wwise" in generators:
        if verbose:
            print("--> Generating Wwise plugin")
        results["c2wwise"] = c2wwise.c2wwise.compile(
            c_src_dir=c_src_dir,
            out_dir=os.path.join(out_dir, "wwise"),
            patch_name=patch_name,
            num_input_channels=num_input_channels,
            num_output_channels=num_output_channels,
            externs=externs,
            verbose=verbose)

    return results


def main():
    tick = time.time()

    parser = argparse.ArgumentParser(
        description="This is the Enzien Audio Heavy compiler. It compiles supported dataflow languages into C,"
                    " and other supported frameworks.")
    parser.add_argument(
        "in_path",
        help="The input dataflow file.")
    parser.add_argument(
        "-o",
        "--out_dir",
        help="Build output path.")
    parser.add_argument(
        "-p",
        "--search_paths",
        nargs="+",
        help="Add a list of directories to search through for abstractions.")
    parser.add_argument(
        "-n",
        "--name",
        default="heavy",
        help="Provides a name for the generated Heavy context.")
    parser.add_argument(
        "-g",
        "--gen",
        nargs="+",
        default=["c"],
        help="List of generator outputs: c, unity, wwise, js, pdext, dpf, fabric")
    parser.add_argument(
        "--results_path",
        help="Write results dictionary to the given path as a JSON-formatted string."
             " Target directory will be created if it does not exist.")
    parser.add_argument(
        "-v",
        "--verbose",
        help="Show debugging information.",
        action="count")
    parser.add_argument(
        "--copyright",
        help="A string indicating the owner of the copyright.")
    args = parser.parse_args()

    in_path = os.path.abspath(args.in_path)
    results = compile_dataflow(
        in_path=in_path,
        out_dir=args.out_dir or os.path.dirname(in_path),
        patch_name=args.name,
        search_paths=args.search_paths,
        generators=args.gen,
        verbose=args.verbose,
        copyright=args.copyright)

    for r in list(results.values()):
        # print any errors
        if r["notifs"].get("has_error", False):
            for i, error in enumerate(r["notifs"].get("errors", [])):
                print("{4:3d}) {2}Error{3} {0}: {1}".format(
                    r["stage"], error["message"], Colours.red, Colours.end, i + 1))

            # only print exception if no errors are indicated
            if len(r["notifs"].get("errors", [])) == 0 and r["notifs"].get("exception", None) is not None:
                print("{2}Error{3} {0} exception: {1}".format(
                    r["stage"], r["notifs"]["exception"], Colours.red, Colours.end))

            # clear any exceptions such that results can be JSONified if necessary
            r["notifs"]["exception"] = []

        # print any warnings
        for i, warning in enumerate(r["notifs"].get("warnings", [])):
            print("{4:3d}) {2}Warning{3} {0}: {1}".format(
                r["stage"], warning["message"], Colours.yellow, Colours.end, i + 1))

    if args.results_path:
        results_path = os.path.realpath(os.path.abspath(args.results_path))
        results_dir = os.path.dirname(results_path)

        if not os.path.exists(results_dir):
            os.makedirs(results_dir)

        with open(results_path, "w") as f:
            json.dump(results, f)

    if args.verbose:
        print("Total compile time: {0:.2f}ms".format(1000 * (time.time() - tick)))


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .HeavyObject import HeavyObject


class ControlPrint(HeavyObject):
    """Prints the first value in a message to the console"""

    preamble = "cPrint"

    @classmethod
    def get_C_header_set(clazz):
        return {"HvControlPrint.h"}

    @classmethod
    def get_C_file_set(clazz):
        return {"HvControlPrint.h", "HvControlPrint.c"}

    @classmethod
    def get_C_init(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_free(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_onMessage(clazz, obj_type, obj_id, inlet_index, args):
        return ["cPrint_onMessage(_c, m, \"{0}\");".format(args["label"])]

    @classmethod
    def get_C_decl(clazz, obj_type, obj_id, args):
        return []

    @classmethod
    def get_C_impl(clazz, obj_type, obj_id, on_message_list, obj_class_dict):
        return []
